/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */
#ifndef LCOM_H
#define LCOM_H

/*
	$Id: lcom.h,v 1.3 2003/03/05 23:47:56 mawag Exp $
	ipc parameters.
*/

/**	communication parameters: client request */
#define LCOM_RQS_SID      900   /* session id */
#define LCOM_RQS_SER      901   /* serial no of request */
#define LCOM_RQS_TYPE     902   /* request type */
#define LCOM_RQS_DBN      903   /* db name */
#define LCOM_RQS_TMS      904   /* server timestamp */
#define LCOM_RQS_ROWID    910   /* rowid (read, delete, update) */
#define LCOM_RQS_QMOD     911   /* mode (query) */
#define LCOM_RQS_SKIP     912   /* skip (query) */
#define LCOM_RQS_SIZE     913   /* output size (query) */
#define LCOM_RQS_QFLG     914   /* flags (query) */
#define LCOM_RQS_KEY      915   /* key value (query) */
#define LCOM_RQS_REC      920   /* data structure (insert, update) */
#define LCOM_RQS_IDX      921   /* index record (insert, update) */

/**	communication parameters: server response */
#define LCOM_RSP_SID      950   /* session id */
#define LCOM_RSP_SER      951   /* serial no of request */
#define LCOM_RSP_DBID     952   /* db id in case of local db */
#define LCOM_RSP_DBN      953   /* db name */
#define LCOM_RSP_TMS      954   /* server timestamp */
#define LCOM_RSP_ERR      955   /* error code */
#define LCOM_RSP_MSG      956   /* error message */
#define LCOM_RSP_NUMT     957   /* total number of records (query, read) */
#define LCOM_RSP_NUMR     958   /* number in this response (query, read) */
#define LCOM_RSP_CERR     959   /* client side error */
#define LCOM_RSP_ROWID    960   /* rowid (query, maxrowid, insert) */
#define LCOM_RSP_DATA     970   /* structure (query, read, open/fdt) */

/** request types. */
#define LCOM_RQST_OPEN        1  /* open db */
#define LCOM_RQST_CREA        2  /* open or create db */
#define LCOM_RQST_CLOS        3  /* close db */
#define LCOM_RQST_MROW       10  /* get maxrowid of db */
#define LCOM_RQST_QRY        11  /* exec query on db */
#define LCOM_RQST_READ       12  /* get rec for rowid */
#define LCOM_RQST_INS        20  /* insert rec */
#define LCOM_RQST_UPD        21  /* update rec */
#define LCOM_RQST_DEL        22  /* delete rec */

/** fdt of request record. */
extern const OpenIsisFdt *openIsisComRqsPar;

/** fdt of response record. */
extern const OpenIsisFdt *openIsisComRspPar;

/** max length of error messages. */
#define LCOM_ERRMSGLEN   1024

/** additional flags in query request. */
#define LCOM_QRYF_READREC   0x01  /* fetch records */

/** max length of key in query request. */
#define LCOM_QRY_KEYLEN   256

/**	callback for delivering response record. */
typedef int DlvCb (OpenIsisSession ses, OpenIsisRec **rsp, void *cld);

/**	ldsp.c: server side request processing */
extern int ldsp_process (OpenIsisSession ses,
	OpenIsisRec *rqs, DlvCb *cb, void *cld);

#endif /* LCOM_H */

