/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */

/*
	$Id: ldsp.c,v 1.13 2003/03/17 09:05:25 mawag Exp $
	server side request processing.
*/

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>  /* vsnprintf */
#include <string.h>

#include "openisis.h"
#include "loi.h"
#include "lcom.h"
#include "ldb.h"
#include "lfdt.h"
#include "luti.h"

#ifdef WIN32
#define vsnprintf _vsnprintf
#endif

#define ISID   0
#define ISER   1
#define ITMS   2
#define SSTS   3

#define RSPVAR                                                 \
	Rec *rsp = 0;                                              \
	Db  *dbh = 0;                                              \
	int sdbid;

#define RSPHEAD                                                \
	rsp = rAddI (ses, rsp, LCOM_RSP_SID, sst[ISID], !0);       \
	rsp = rAddI (ses, rsp, LCOM_RSP_SER, sst[ISER], !0);

#define RSPNOERR                                               \
	rsp = rAddI (ses, rsp, LCOM_RSP_ERR, 0, !0);

#define RSPDB                                                  \
	RADDS (ses, rsp, LCOM_RSP_DBN, dbn, !0);

#define RSPFREE                                                \
	if (rsp) { mFree (ses, rsp); }

#define GETDB(rname)                                           \
	sdbid = ldspGetDb (&dbh, sst, dbn);                        \
	if (0 > sdbid) {                                           \
		int rt = _RqsError (ses, sst, ERR_INVAL, cb,           \
			cld, rname, sst[ITMS] ?                            \
			"db %s remounted" : "db %s not open", dbn);        \
		RSPFREE                                                \
		return rt;                                             \
	}

#define DELIVER                                                \
	if (rsp) {                                                 \
		int rt = (*cb) (ses, &rsp, cld);                       \
		RSPFREE                                                \
		return rt;                                             \
	}                                                          \
	sMsg (ses, LOG_ERROR, "ldsp: deliver: rsp == 0");          \
	return 0;

/* ************************************************************
*/

static int ldspGetDb (Db **db, int sst[], const char *dbn) {
	Db *dbh = nDbByName (stub0, dbn);
	if (! dbh) {
		sst[ITMS] = 0;
		return -1;
	}
	if (sst[ITMS]) {
		if (-1 == sst[ITMS] || /* dont want default of rInt call */
			(dbh->tms && dbh->tms != sst[ITMS])
		) {
			sst[ITMS] = dbh->tms;
			return -1;
		}
	}
	if (db) {
		*db = dbh;
	}
	return dbh->dbid;
}

static int _RqsError (
	Session ses, int sst[], int err, DlvCb *cb,
	void *cld, const char *rname, const char *fmt, ...
) {
	RSPVAR
	char msg[LCOM_ERRMSGLEN];
	va_list ap;
	(void)sdbid; (void)dbh;
	RSPHEAD
	if (sst[ITMS] && -1 != sst[ITMS]) {
		rsp = rAddI (ses, rsp, LCOM_RSP_TMS, sst[ITMS], !0);
	}
	rsp = rAddI (ses, rsp, LCOM_RSP_ERR, err, !0);
	if (0 != fmt && 0 != *fmt) {
		va_start (ap, fmt);
		vsnprintf (msg, LCOM_ERRMSGLEN - 1, fmt, ap);
		msg[LCOM_ERRMSGLEN - 1] = 0;
		va_end (ap);
		RADDS (ses, rsp, LCOM_RSP_MSG, msg, !0);
		sMsg (ses, LOG_WARN, "ldsp: err on rqs %s: %d(%x) - %s",
			rname, err, err, msg);
	}
	else {
		sMsg (ses, LOG_WARN, "ldsp: err on rqs %s: %d(%x)",
			rname, err, err);
	}
	DELIVER
}

static int _RqsOpen (
	Session ses, int sst[], const char *dbn, DlvCb *cb, void *cld
) {
#define STRERRNOLEN  128
	RSPVAR
	Schema *sc;
	char    msg[STRERRNOLEN];
	sst[ITMS] = 0;
	sc = nSchema (stub0);
	if (! sc) {
		int rt = _RqsError (ses, sst, ERR_BADF, cb, cld,
			"open", "%s: local schema not initialized", dbn);
		RSPFREE
		return rt;
	}
	errno = 0;
	dbh = cDOpen (dbn, 0, sc->cfg);
	if (! dbh) {
		int rt;
		strncpy (msg, strerror (errno), STRERRNOLEN - 1)
			[STRERRNOLEN - 1] = 0;
		rt = _RqsError (ses, sst, ERR_BADF, cb, cld,
			"open", "%s: %s", dbn, msg);
		RSPFREE
		return rt;
	}
	sdbid = dbh->dbid;
	RSPHEAD
	RSPNOERR
	RSPDB
	rsp = rAddI (ses, rsp, LCOM_RSP_DBID, dbh->dbid, !0);
	rsp = rAddI (ses, rsp, LCOM_RSP_TMS, dbh->tms, !0);
	if (dbh->fdt) {
		rsp = rAddI (ses, rsp, LCOM_RSP_NUMT, dbh->fdt->len, !0);
		rsp = rAddI (ses, rsp, LCOM_RSP_NUMR, dbh->fdt->len, !0);
		rsp = fdt2rec (ses, rsp, dbh->fdt, LCOM_RSP_DATA);
	}
	else {
		rsp = rAddI (ses, rsp, LCOM_RSP_NUMT, 0, !0);
		rsp = rAddI (ses, rsp, LCOM_RSP_NUMR, 0, !0);
	}
	DELIVER
}

static int _RqsClose (
	Session ses, int sst[], const char *dbn, DlvCb *cb, void *cld
) {
	RSPVAR
	int err;
	GETDB("close")
	err = cDClose (sdbid);
	RSPHEAD
	rsp = rAddI (ses, rsp, LCOM_RSP_DBID, sdbid, !0);
	rsp = rAddI (ses, rsp, LCOM_RSP_ERR, err, !0);
	RSPDB
	DELIVER
}

static int _RqsMaxrow (
	Session ses, int sst[], const char *dbn, DlvCb *cb, void *cld
) {
	RSPVAR
	int maxrow;
	GETDB("maxrow")
	maxrow = dMaxId (ses, sdbid);
	if (0 > maxrow) {
		int rt = _RqsError (ses, sst, ERR_TRASH, cb, cld, "maxrow",
			"no maxrow for %s", dbn);
		RSPFREE
		return rt;
	}
	RSPHEAD
	RSPNOERR
	RSPDB
	rsp = rAddI (ses, rsp, LCOM_RSP_ROWID, maxrow, !0);
	DELIVER
}

static int _RqsQuery (
	Session ses, int sst[], const char *dbn, const char *key,
	int mode, int skip, int size, int flags, DlvCb *cb, void *cld
) {
	RSPVAR
	Set         set;
	Rec        *rec;
	int        lrt;
	int         j;
	int         rdr = LCOM_QRYF_READREC & flags;
	GETDB("query")
	set.len = 0; /* input */
	lrt = dQuery (ses, &set, sdbid, key, mode, skip);
	if (0 > lrt) {
		int rt = _RqsError (ses, sst, ERR_TRASH, cb, cld, "query",
			"%d,%s = %d", sdbid, key, lrt);
		RSPFREE
		return rt;
	}
	RSPHEAD
	RSPNOERR
	RSPDB
	rsp = rAddI (ses, rsp, LCOM_RSP_NUMT, lrt, !0);
	if (0 < size && lrt > size) {
		lrt = size;
	}
	rsp = rAddI (ses, rsp, LCOM_RSP_NUMR, lrt, !0);
	for (j = 0; lrt > j; ++j) {
		rsp = rAddI (ses, rsp, LCOM_RSP_ROWID, set.id[j], !0);
	}
	if (rdr) {
		for (j = 0; rsp && lrt > j; ++j) {
			rec = dRead (ses, sdbid, set.id[j]);
			if (0 == rec) {
				int rt = _RqsError (ses, sst, ERR_TRASH, cb, cld,
					"query", "non existent rec %d", set.id[j]);
				RSPFREE
				return rt;
			}
			rsp = luti_wrap (ses, rsp, rec, LCOM_RSP_DATA);
			mFree (ses, rec);
		}
	}
	DELIVER
}

static int _RqsRead (Session ses,
	int sst[], const char *dbn, int row, DlvCb *cb, void *cld
) {
	RSPVAR
	Rec *rec;
	GETDB("read")
	RSPHEAD
	RSPNOERR
	RSPDB
	rsp = rAddI (ses, rsp, LCOM_RSP_ROWID, row, !0);
	rec = dRead (ses, sdbid, row);
	if (rec) {
		rsp = rAddI (ses, rsp, LCOM_RSP_NUMT, 1, !0);
		rsp = rAddI (ses, rsp, LCOM_RSP_NUMR, 1, !0);
		rsp = luti_wrap (ses, rsp, rec, LCOM_RSP_DATA);
		mFree (ses, rec);
	}
	else {
		rsp = rAddI (ses, rsp, LCOM_RSP_NUMT, 0, !0);
		rsp = rAddI (ses, rsp, LCOM_RSP_NUMR, 0, !0);
	}
	DELIVER
}

static int _RqsUpdate (Session ses, int sst[],
	const char *dbn, int row, Rec *rec, Rec *idx, DlvCb *cb, void *cld
) {
	RSPVAR
	int rtw;
	GETDB("update")
	rec->rowid = row;
	rec->dbid = sdbid;
	rtw = dWritex (ses, sdbid, rec, idx);
	if (rtw) {
		rtw = _RqsError (ses, sst, ERR_TRASH, cb, cld,
			"update", "write error %d", rtw);
		RSPFREE
		if (rec) { mFree (ses, rec); }
		if (idx) { mFree (ses, idx); }
		return rtw;
	}
	RSPHEAD 
	RSPNOERR
	RSPDB
	rsp = rAddI (ses, rsp, LCOM_RSP_ROWID, rec->rowid, !0);
	if (rec) { mFree (ses, rec); }
	if (idx) { mFree (ses, idx); }
	DELIVER
}

static int _RqsDelete (Session ses,
	int sst[], const char *dbn, int row, DlvCb *cb, void *cld
) {
	(void)dbn; (void)row;
	return _RqsError (ses, sst, ERR_TRASH, cb, cld, "delete",
		"request not implemented yet");
}

/* ************************************************************
	package functions
*/

int ldsp_process (Session ses, Rec *rqs, DlvCb *cb, void *cld) {
	char   buf[DBNLEN];
	char  *dbn;
	int    sst[SSTS];
	int    rtyp;
	if (0 == ses) {
		sMsg (ses0, LOG_ERROR, "ldsp_process: no session");
		return -1;
	}
	if (0 == rqs) {
		sMsg (ses, LOG_ERROR, "ldsp_process: no request");
		return -1;
	}
	if (0 == cb) {
		sMsg (ses, LOG_ERROR, "ldsp_process: no sender");
		return -1;
	}
	rtyp      = rInt (rqs, LCOM_RQS_TYPE, -1, 0);
	sst[ISID] = rInt (rqs, LCOM_RQS_SID, -1, 0);
	sst[ISER] = rInt (rqs, LCOM_RQS_SER, -1, 0);
	sst[ITMS] = rInt (rqs, LCOM_RQS_TMS, -1, 0);
	dbn       = rString (rqs, LCOM_RQS_DBN, 0, buf, DBNLEN);
	if (!dbn) {
		i2a (buf, rtyp);
		return _RqsError (ses, sst, ERR_INVAL, cb, cld, buf,
			"no db name given");
	}
	switch (rtyp) {
	case LCOM_RQST_OPEN:
	case LCOM_RQST_CREA:
		return _RqsOpen (ses, sst, dbn, cb, cld);
	case LCOM_RQST_CLOS:
		return _RqsClose (ses, sst, dbn, cb, cld);
	case LCOM_RQST_MROW:
		return _RqsMaxrow (ses, sst, dbn, cb, cld);
	case LCOM_RQST_QRY: {
		char  kybuf[LCOM_QRY_KEYLEN];
		char *key;
		int  mode, skip, size;
		int   flg;
		key  = rString (rqs, LCOM_RQS_KEY, 0, kybuf, LCOM_QRY_KEYLEN);
		mode = rInt (rqs, LCOM_RQS_QMOD, -1, 0);
		skip = rInt (rqs, LCOM_RQS_SKIP, 0, 0);
		size = rInt (rqs, LCOM_RQS_SIZE, -1, 0);
		flg  = rInt (rqs, LCOM_RQS_QFLG, 0, 0);
		if (0 > mode || 0 == key) {
			return _RqsError (ses, sst, ERR_INVAL, cb, cld, "query",
				"missing mode or key");
		}
		return _RqsQuery (ses, sst, dbn, key,
			mode, skip, size, flg, cb, cld);
	}
	case LCOM_RQST_READ: {
		int rowid;
		rowid = rInt (rqs, LCOM_RQS_ROWID, -1, 0);
		if (0 > rowid) {
			return _RqsError (ses, sst, ERR_INVAL, cb, cld, "read",
				"missing rowid");
		}
		return _RqsRead (ses, sst, dbn, rowid, cb, cld);
	}
	case LCOM_RQST_INS:
	case LCOM_RQST_UPD: {
		Rec  *rec, *idx;
		int  rowid = 0;
		int   ins = rtyp == LCOM_RQST_INS;
		rec = luti_unwrap (ses, rqs, 0, LCOM_RQS_REC, -1);
		idx = luti_unwrap (ses, rqs, 0, LCOM_RQS_IDX, -1);
		if (!(rec || idx)) {
			return _RqsError (ses, sst, ERR_INVAL, cb, cld,
				"update", "missing record or index");
		}
		if (! ins) {
			rowid = rInt (rqs, LCOM_RQS_ROWID, 0, 0);
			if (0 >= rowid) {
				return _RqsError (ses, sst, ERR_INVAL, cb, cld,
					"update", "missing rowid");
			}
		}
		return _RqsUpdate (ses, sst, dbn, rowid, rec, idx, cb, cld);
	}
	case LCOM_RQST_DEL: {
		int rowid;
		rowid = rInt (rqs, LCOM_RQS_ROWID, -1, 0);
		if (0 > rowid) {
			return _RqsError (ses, sst, ERR_INVAL, cb, cld, "delete",
				"missing rowid");
		}
		return _RqsDelete (ses, sst, dbn, rowid, cb, cld);
	}
	}
	i2a (buf, rtyp);
	return _RqsError (ses, sst, ERR_INVAL, cb, cld, buf,
		"unrecognized request type");
}

