/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2001-2002 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */

/*
	$Id: lfdt.c,v 1.8 2003/03/16 18:58:49 mawag Exp $
	implementation of FDT calls and static FDTs.
*/

#include <stdio.h>
#include <string.h>

#include "openisis.h"
#include "lfdt.h"
#include "luti.h"

#define EXT_FDT      ".fdt"
#define FD_NAMELEN   30
#define FD_PATLEN    20
#define FD_TOOLLEN   6
#define FD_MAXTOOLS  64


/* ************************************************************
	private types
*/
/* ************************************************************
	private data
*/
static Fd paramf[] = {
/* i, t, r, l, n, s, d */
	/* 700 ... OpenIsis system parameters */
	{ OPENISIS_SPATH, FTX, 0, 256, "syspath", "", "Gloval Database Path" },
	{ OPENISIS_SLOGF, FTX, 0, 256, "logfile", "", "Logfile Name" },
	{ OPENISIS_SLOGV, FTE, 0, 256, "v", "", "Verbosity of Logging" },
	{ OPENISIS_SLOGV, FTV, 0, 0, "off", "", "don't log anything" },
	{ OPENISIS_SLOGV, FTV, 0, 1, "fatal", "", "" },
	{ OPENISIS_SLOGV, FTV, 0, 3, "syserr", "", "" },
	{ OPENISIS_SLOGV, FTV, 0, 4, "error", "", "" },
	{ OPENISIS_SLOGV, FTV, 0, 5, "warn", "", "" },
	{ OPENISIS_SLOGV, FTV, 0, 6, "info", "", "" },
	{ OPENISIS_SLOGV, FTV, 0, 7, "verbose", "", "" },
	{ OPENISIS_SLOGV, FTV, 0, 8, "trace", "", "" },
	{ OPENISIS_SLOGV, FTV, 0, 9, "debug", "", "" },
	{ OPENISIS_SLOGV, FTV, 0, 10, "all", "", "" },
	{ OPENISIS_SSCHEMA, FTX, 1, 256, "schema", "shp", "Schema entry" },
	/* db part starts at pos 14, see below */
	{ OPENISIS_DNAME, FTX, 0, 256, "db", "", "Database Name" },
	{ OPENISIS_DTYPE, FTE, 0, 256, "format", "", "Database Format" },
	{ OPENISIS_DTYPE, FTV, 0, 0, "autoformat", "", "Database Format" },
	{ OPENISIS_DTYPE, FTV, 0, 1, "naligned", "", "Database Format" },
	{ OPENISIS_DTYPE, FTV, 0, 2, "aligned", "", "Database Format" },
	{ OPENISIS_DRO  , FTB, 0, 1, "ro", "", "Readonly Flag" },
	{ OPENISIS_DPATH, FTX, 0, 256, "dbpath", "", "Database Path" },
	{ OPENISIS_DDUMP, FTB, 0, 1, "internaldump", "", "Internal Dump Flag" }
};
static Fdt paramS = {
	sizeof( paramf )/sizeof( paramf[0] ),
	paramf,
	0
};
static Fdt paramD = {
	sizeof( paramf )/sizeof( paramf[0] ) - 14,
	paramf + 14,
	0
};

/* ************************************************************
	public data
*/
const Fdt *openIsisSyspar = &paramS;
const Fdt *openIsisDbpar = &paramD;


/* ************************************************************
	private functions
*/

static char *fgetsnl (FILE *fp) {
	static char buf[4096];
	char *res;
	int len;
	while (1) {
		if (0 == (res = fgets (buf, 4095, fp))) {
			break;
		}
		buf[4095] = 0;
		if (0 == (len = strlen (res))) {
			continue;
		}
		if ('\n' == res[--len]) {
			if (0 == len) {
				continue;
			}
			res[len] = 0;
		}
		break;
	}
	return res;
}

static void strrtrim (char *str, int len) {
	char *p = str + len - 1;
	while (p >= str && (' ' == *p || '\t' == *p)) {
		*p-- = 0;
	}
}

static char *desc2name (char *d, char *n) {
	char *name = n;
	while (1) {
		if (!(*n = *d)) {
			return name;
		}
		if ('A' <= *n && 'Z' >= *n) {
			*n += 'a' - 'A';
			++n;
			++d;
			continue;
		}
		if ('_' == *n ||
			('a' <= *n && 'z' >= *n) ||
			('0' <= *n && '9' >= *n)) {
			++n;
			++d;
			continue;
		}
		while (1) {
			++d;
			if (!*d) {
				*n = 0;
				return name;
			}
			if (('A' <= *d && 'Z' >= *d) ||
				('a' <= *d && 'z' >= *d) ||
				('0' <= *d && '9' >= *d)) {
				break;
			}
		}
		*n++ = '_';
	}
}

static int line2Fd (char *line,
	char *name, char *pat, int *tag, int *len, int *type, int *rep
) {
	if (FD_NAMELEN + FD_PATLEN >= strlen (line)) {
		return 0;
	}
	strncpy (name, line, FD_NAMELEN) [FD_NAMELEN] = 0;
	strrtrim (name, strlen (name));
	if (! *name) {
		return 0;
	}
	line += FD_NAMELEN;
	strncpy (pat, line, FD_PATLEN) [FD_PATLEN] = 0;
	strrtrim (pat, strlen (pat));
	line += FD_PATLEN;
	if (4 != sscanf (line, "%d %d %d %d", tag, len, type, rep)) {
		return 0;
	}
	switch (*type) {
	case 0: *type = OPENISIS_FTX; break;
	case 1: *type = OPENISIS_FTA; break;
	case 2: *type = OPENISIS_FTN; break;
	case 3: *type = OPENISIS_FTP; break;
	default: return 0;
	}
	if (0 != *rep && 1 != *rep) {
		return 0;
	}
	return !0;
}

static int line2Tool (char *line, int *type, char **tools) {
	int   len;
	int   num = 0;
	switch (*line) {
	case 'w':
	case 'W':
		*type = OPENISIS_DFMT;
		break;
	case 'f':
	case 'F':
		*type = OPENISIS_DPFT;
		break;
	case 's':
	case 'S':
		*type = OPENISIS_DFST;
		break;
	default:
		return 0;
	}
	line += 2;
	len = strlen (line);
	while (0 < len && FD_MAXTOOLS > num) {
		tools[num] = line;
		if (FD_TOOLLEN > len) {
			strrtrim (line, len);
			++num;
			break;
		}
		strrtrim (line, FD_TOOLLEN);
		line += FD_TOOLLEN;
		len -= FD_TOOLLEN;
		++num;
	}
	return num;
}

/* ************************************************************
	package functions
*/

#if 0

const char *lfdt_argv (Fdt *fdt, int fdid,
	const char *name, const char *val, const char *dflt) {
	Fd *D;
	if (! fdt || ! name || ! val) {
		return dflt;
	}
	D = fdById (fdt, fdid);
	if (! D) {
		return dflt;
	}
	if (strcmp (name, D->name)) {
		return dflt;
	}
	return val;
}

int lfdt_argi (Fdt *fdt, int fdid,
	const char *name, const char *val, int dflt) {
	val = lfdt_argv (fdt, fdid, name, val, 0);
	if (! val) {
		return dflt;
	}
	return a2id (val, -1, dflt);
}

int lfdt_arge (Fdt *fdt, int fdid,
	const char *name, const char *val, int dflt) {
	Fd  *D;
	int  ev;
	if (! fdt || ! name) {
		return NOENUM;
	}
	D = fdById (fdt, fdid);
	if (! D) {
		return NOENUM;
	}
	if (strcmp (name, D->name)) {
		return NOENUM;
	}
	if (! val) {
		return dflt;
	}
	ev = fdEnum (fdt, D->id, val);
	if (NOENUM == ev) {
		return dflt;
	}
	return ev;
}

#endif /* 0 */

/* ************************************************************
	public functions
*/

Fd* fdById ( const Fdt *fdt, int id )
{
	Fd *f = fdt->fd, *e = f + fdt->len;
	for ( ; f < e; f++ )
		if ( id == f->id )
			return f;
	return 0;
}	/* fdById */


Fd* fdByName ( const Fdt *fdt, const char *name )
{
	Fd *f = fdt->fd, *e = f + fdt->len;
	const char *p;
	int l;
	if ( '-' == *name ) /* ignore leading dash */
		name++;
	if ( '0' <= *name && *name <= '9' )
		return fdById( fdt, a2i( name, -1 ) );
	p = name;
	while ( 'a' <= *p ? *p <= 'z' : '9' >= *p ? *p >= '0' : '_' == *p )
		p++;
	l = p - name;
	if ( ! l || l > OPENISIS_FD_NAMELEN - 1 )
		return 0;
	for ( ; f < e; f++ )
		if ( *name == *f->name
			&& !(0xf0 & f->type) /* is field */
			&& !memcmp( name, f->name, l )
			&& !f->name[l]
		)
			return f;
	return 0;
}	/* fdByName */


/**
	lookup enum:
	if name is numeric, the value is returned, if legal
	considered are value entries with given id or id 0
	- if there is an exact match with same id, this is used
	- if there is an exact match with id 0, this is used
	-	if name is a unique prefix on given id, this is used
	-	if name is no prefix on given id, but a unique prefix on 0, this is used
*/
int fdEnum ( Fdt *fdt, int id, const char *name )
{
	Fd *f = fdt->fd, *e = f + fdt->len,
		*x0 = 0, *pi = 0, *p0 = 0;
	int ui = 1, u0 = 1; /* unique */
	int l = strlen( name );
	if ( ! l || l > OPENISIS_FD_NAMELEN - 1 )
		return NOENUM;
	if ( ('0' <= *name && *name <= '9')
		|| ('-' == *name && '0' <= name[1] && name[1] <= '9')
	) {
		int v = a2i( name, l );
		for ( ; f < e; f++ ) {
			if ( FTV == f->type && v == f->len && (f->id == id || !f->id) )
				return v;
		}
		return NOENUM;
	}
	for ( ; f < e; f++ ) {
		if ( FTV != f->type
			|| *name != *f->name
			|| memcmp( name, f->name, l )
			|| (f->id && f->id != id)
		)
			continue;
		if ( !f->name[l] ) { /* exact match */
			if ( f->id == id )
				return f->len;
			x0 = f; /* f->id is 0 */
			continue;
		}
		/* prefix match */
		if ( f->id == id ) {
			if ( pi )
				ui = 0;
			else
				pi = f;
		} else
			if ( p0 )
				u0 = 0;
			else
				p0 = f;
	}
	return x0 ? x0->len
		: pi ? (ui ? pi->len : NOENUM)
		: p0 && u0 ? p0->len : NOENUM;
}	/* fdEnum */


Fdt *cFdt ( Session ses, int len, Fd *fd, Rec *rec )
{
	Fdt *fdt = (Fdt*) mAlloc (ses, sizeof(Fdt));
	if ( ! fdt )
		return 0;
	fdt->len = len;
	fdt->fd = fd;
	fdt->rec = rec;
	/* TODO: if we have no fd, create one from rec */
	return fdt;
}	/* cFdt */

void cFreeFdt ( Session ses, Fdt *fdt ) {
	if (fdt) {
		if (fdt->fd) {
			mFree (ses, fdt->fd);
		}
		if (fdt->rec) {
			mFree (ses, fdt->rec);
		}
		mFree (ses, fdt);
	}
}


#define CLN_FR                                      \
	if (0 != fds && 0 != *fds) {                    \
		mFree (ses, *fds);                          \
		*fds = 0;                                   \
	}                                               \
	if (0 != rec && 0 != *rec) {                    \
		mFree (ses, *rec);                          \
		*rec = 0;                                   \
	}                                               \
	return -1


int fdtFromFile (char *path, Fd **fds, Rec **rec) {
	Session ses;
	char    name[1 + FD_NAMELEN];
	char    pat[1 + FD_PATLEN];
	char    tool[1 + FD_TOOLLEN];
	char   *tools[FD_MAXTOOLS];
	char   *line;
	Fd     *nfd;
	FILE   *fp;
	int     tag, len, num, type, rep, j;
	ses = OPENISIS_SES0 ();
	/* ldb::setext */
	len = strlen (path) - 4;
	memcpy (path + len, EXT_FDT, 4);
	if ( 'A'<=path[len-1] && path[len-1]<= 'Z' ) {
		char *p = path + len;
		for ( ;*p; p++ ) /* use uppercase extensions */
			if ( 'a' <= *p && *p <= 'z' )
				*p -= 'a'-'A';
	}
	if (0 != fds) {
		*fds = 0;
	}
	if (0 != rec) {
		*rec = 0;
	}
	fp = fopen (path, "r");
	if (0 == fp) {
		sMsg (ses, LOG_INFO, "no such fdt: %s", path);
		return -1;
	}
	sMsg (ses, LOG_VERBOSE, "> reading fdt: %s", path);
	num = 0;
	while (0 != (line = fgetsnl (fp))) {
		if ('*' == *line) {
			continue;
		}
		if (':' == line[1]) {
			if (0 != rec) {
				len = line2Tool (line, &type, tools);
				for (j = 0; len > j; ++j) {
					strncpy (tool, tools[j], FD_TOOLLEN) [FD_TOOLLEN] = 0;
					OPENISIS_RADDS (ses, *rec, type, tool, !0);
					if (0 == *rec) {
						CLN_FR;
					}
				}
			}
			continue;
		}
		if (0 != fds &&
			line2Fd (line, name, pat, &tag, &len, &type, &rep)) {
			nfd = mAlloc (ses, (1 + num) * sizeof (Fd));
			if (0 == nfd) {
				CLN_FR;
			}
			if (0 != num) {
				memcpy (nfd, *fds, num * sizeof (Fd));
				mFree (ses, *fds);
			}
			*fds = nfd;
			nfd = *fds + num++;
			memset (nfd, 0, sizeof (Fd));
			nfd->id = tag;
			nfd->type = (char) type;
			nfd->rep = (char) rep;
			nfd->len = (short) len;
			desc2name (name, nfd->name);
			strcpy (nfd->subf, pat);
			strcpy (nfd->desc, name);
		}
	}
	fclose (fp);
	sMsg (ses, LOG_VERBOSE, "< %d entries in fdt", num);
	return num;
}

Rec *fd2rec (Session ses, Fd *fd) {
	char buf[12];
	Rec *rec = 0;
	if (0 == fd) {
		return 0;
	}
	i2a (buf, fd->id);
	RADDS (ses, rec, OPENISIS_FDID, buf, !0);
	i2a (buf, fd->type);
	RADDS (ses, rec, OPENISIS_FDTYPE, buf, !0);
	i2a (buf, fd->rep);
	RADDS (ses, rec, OPENISIS_FDREP, buf, !0);
	i2a (buf, fd->len);
	RADDS (ses, rec, OPENISIS_FDLEN, buf, !0);
	RADDS (ses, rec, OPENISIS_FDNAME, fd->name, !0);
	RADDS (ses, rec, OPENISIS_FDSUBF, fd->subf, !0);
	RADDS (ses, rec, OPENISIS_FDDESC, fd->desc, !0);
	return rec;
}

Rec *fdt2rec (Session ses, Rec *rec, Fdt *fdt, int tag) {
	Fd   *F;
	Rec  *R;
	if (! fdt) {
		return rec;
	}
	for (F = fdt->fd; fdt->len > F - fdt->fd; ++F) {
		R = fd2rec (ses, F);
		if (!R) {
			return 0;
		}
		rec = luti_wrap (ses, rec, R, tag);
		mFree (ses, R);
		if (!rec) {
			return 0;
		}
	}
	return rec;
}

#define CLN_F                                       \
	if (0 == buf) {                                 \
		mFree (ses, fd);                            \
	}                                               \
	return 0

Fd *rec2fd (Session ses, Rec *rec, Fd *buf) {
	Field   *F;
	Fd      *fd;
	int      val, len;
	int      got = 0;
	if (0 == rec) {
		return 0;
	}
	if (0 == buf) {
		fd = mAlloc (ses, sizeof (Fd));
		if (0 == fd) {
			return 0;
		}
	}
	else {
		fd = buf;
	}
	memset (fd, 0, sizeof (Fd));
	for (F = rec->field; F - rec->field < rec->len; ++F) {
		switch (F->tag) {
		case OPENISIS_FDID:
			if (0 != (0x01 & got)) { CLN_F; } got |= 0x01;
			if (0 > (val = a2id (F->val, (int)F->len, -1))) { CLN_F; }
			fd->id = val;
			break;
		case OPENISIS_FDTYPE:
			if (0 != (0x02 & got)) { CLN_F; } got |= 0x02;
			if (0 > (val = a2id (F->val, (int)F->len, -1))) { CLN_F; }
			fd->type = (char) val;
			break;
		case OPENISIS_FDREP:
			if (0 != (0x04 & got)) { CLN_F; } got |= 0x04;
			if (0 > (val = a2id (F->val, (int)F->len, -1))) { CLN_F; }
			fd->rep = (char)(0 != val);
			break;
		case OPENISIS_FDLEN:
			if (0 != (0x08 & got)) { CLN_F; } got |= 0x08;
			if (0 > (val = a2id (F->val, (int)F->len, -1))) { CLN_F; }
			fd->len = (short) val;
			break;
		case OPENISIS_FDNAME:
			if (0 != (0x10 & got)) { CLN_F; } got |= 0x10;
			if (! F->len) { CLN_F; }
			if (30 < (len = (int)F->len)) {
				len = 30;
			}
			strncpy (fd->name, F->val, len) [len] = 0;
			break;
		case OPENISIS_FDSUBF:
			if (0 != (0x20 & got)) { CLN_F; } got |= 0x20;
			if (F->len) {
				if (20 < (len = (int)F->len)) {
					len = 20;
				}
				strncpy (fd->subf, F->val, len) [len] = 0;
			}
			break;
		case OPENISIS_FDDESC:
			if (0 != (0x40 & got)) { CLN_F; } got |= 0x40;
			if (F->len) {
				if (30 < (len = (int)F->len)) {
					len = 30;
				}
				strncpy (fd->desc, F->val, len) [len] = 0;
			}
			break;
		}
	}
	if (0x13 != (got & 0x13)) { CLN_F; }
	return fd;
}

Fdt *rec2fdt (Session ses, Rec *rec, int idnum, int idfd, int *pos) {
	Rec     *R;
	Fdt     *fdt;
	Fd       fdbuf;
	Fd      *D, *arr, *E, *A;
	int      num;
	if (!rec) {
		return 0;
	}
	num = rInt (rec, idnum, -1, pos);
	if (0 >= num) {
		return 0;
	}
	arr = (Fd*) mAlloc (ses, num * sizeof (Fd));
	if (!arr) {
		return 0;
	}
	for (E = (A = arr) + num; E > A; ++A) {
		R = luti_unwrap (ses, rec, pos, idfd, -1);
		if (!R) {
			mFree (ses, arr);
			return 0;
		}
		D = rec2fd (ses, R, &fdbuf);
		mFree (ses, R);
		if (!D) {
			mFree (ses, arr);
			return 0;
		}
		memcpy (A, D, sizeof (Fd));
	}
	fdt = cFdt (ses, num, arr, 0);
	if (!fdt) {
		mFree (ses, arr);
		return 0;
	}
	return fdt;
}

