/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2001-2002 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */
#ifndef LOI_H

/*
	$Id: loi.h,v 1.22 2003/03/28 18:13:39 kripke Exp $
	lib openisis: provides the "openisis namespace",
	i.e. short names for most openIsis names,
	so we don't get to dislike those letters ...
*/

#include <assert.h>

#define OPENISIS_NOPRE09
#include "openisis.h"

#define Fld        OpenIsisFld
#define Field      OpenIsisField
#define Raw        OpenIsisRaw
#define Rec        OpenIsisRec
#define BASESZ     OPENISIS_BASESZ
#define INITREC    OPENISIS_INITREC
#define INITSTD    OPENISIS_INITSTD
#define INITBUF    OPENISIS_INITBUF
#define CLRREC     OPENISIS_CLRREC
#define RECOK      OPENISIS_RECOK
#define rGet       openIsisRGet
#define rOccurence openIsisROccurence
#define rInt       openIsisRInt
#define rInt2      openIsisRInt2
#define rEnum      openIsisREnum
#define rString    openIsisRString
#define rString2   openIsisRString2
#define a2i        openIsisA2i
#define a2il       openIsisA2il
#define a2id       openIsisA2id
#define i2a        openIsisI2a
#define u2a        openIsisU2a
#define rSer       openIsisRSerialize
#define rDeser     openIsisRDeserialize

#define FTX        OPENISIS_FTX
#define FTA        OPENISIS_FTA
#define FTN        OPENISIS_FTN
#define FTP        OPENISIS_FTP
#define FTI        OPENISIS_FTI
#define FTE        OPENISIS_FTE
#define FTB        OPENISIS_FTB
#define FTT        OPENISIS_FTT
#define FTS        OPENISIS_FTS
#define FTF        OPENISIS_FTF
#define FTV        OPENISIS_FTV
#define NOENUM     OPENISIS_NOENUM
#define Fd         OpenIsisFd
#define Fdt        OpenIsisFdt
#define fdById     openIsisFdById
#define fdByName   openIsisFdByName
#define fdEnum     openIsisFdEnum

#define Ses        OpenIsisSes
#define Session    OpenIsisSession /* Ses* */
#define mAlloc     openIsisMAlloc
#define mFree      openIsisMFree
#define mDup       openIsisMDup
#define toHtml     openIsisToHtml
#define utf8Chk    openIsisUtf8Chk

#define Stb        OpenIsisStb
#define Stub       OpenIsisStub /* Stb* */
#define Db         OpenIsisDb
#define Schema     OpenIsisSchema
#define DBNLEN     OPENISIS_DB_NAMELEN
#define SCNLEN     OPENISIS_SC_NAMELEN
#define DBIMSK     OPENISIS_DBIDMSK
#define SCIMSK     OPENISIS_SCIDMSK
#define nSchema    openIsisNSchema
#define nSend      openIsisNSend
#define nRecv      openIsisNRecv
#define nClean     openIsisNClean
#define nDbByName  openIsisNDbByName
#define nDbById    openIsisNDbById

#define Ios        OpenIsisIos
#define Stream     OpenIsisStream /* Ios* */
#define SFunc      OpenIsisSFunc
#define ERR_OK     OPENISIS_ERR_OK
#define ERR_EOF    OPENISIS_ERR_EOF
#define ERR_FAULT  OPENISIS_ERR_FAULT
#define ERR_INVAL  OPENISIS_ERR_INVAL
#define ERR_BADF   OPENISIS_ERR_BADF
#define ERR_IO     OPENISIS_ERR_IO
#define ERR_NOMEM  OPENISIS_ERR_NOMEM
#define ERR_BUSY   OPENISIS_ERR_BUSY
#define ERR_TRASH  OPENISIS_ERR_TRASH
#define ERR_IDIOT  OPENISIS_ERR_IDIOT
#define ERR_ERRORS OPENISIS_ERR_ERRORS
#define ERR_MASK   OPENISIS_ERR_MASK
#define ERR_SHIFT  OPENISIS_ERR_SHIFT
#define LOG_OFF    OPENISIS_LOG_OFF
#define LOG_FATAL  OPENISIS_LOG_FATAL
#define LOG_SYSERR OPENISIS_LOG_SYSERR
#define LOG_IOERR  OPENISIS_LOG_IOERR
#define LOG_ERROR  OPENISIS_LOG_ERROR
#define LOG_WARN   OPENISIS_LOG_WARN
#define LOG_INFO   OPENISIS_LOG_INFO
#define LOG_VERBOSE OPENISIS_LOG_VERBOSE
#define LOG_TRACE  OPENISIS_LOG_TRACE
#define LOG_DEBUG  OPENISIS_LOG_DEBUG
#define LOG_ALL    OPENISIS_LOG_ALL
#define LOG_NOCHANGE OPENISIS_LOG_NOCHANGE
#define LOG_LEVELS OPENISIS_LOG_LEVELS
#define LOG_MASK   OPENISIS_LOG_MASK
#define LOG_SHIFT  OPENISIS_LOG_SHIFT
#define MSG_VA     OPENISIS_MSG_VA
#define LogLevel   OpenIsisLogLevel
#define sMsg       openIsisSMsg
#define sOpen      openIsisSOpen
#define sGets      openIsisSGets
#define sReadln    openIsisSReadln
#define sGetr      openIsisSGetr

#define rDup       openIsisRDup
#define rMsg       openIsisRMsg
#define rFmt       openIsisRFmt
#define rSplitf    openIsisRSplitf
#define rSet       openIsisRSet
#define rAddI      openIsisRAddI
#define RSPACE     OPENISIS_RSPACE
#define RADD       OPENISIS_RADD
#define RCAT       OPENISIS_RCAT
#define RADDS      OPENISIS_RADDS
#define RADDF      OPENISIS_RADDF
#define RCATS      OPENISIS_RCATS
#define RARGC      OPENISIS_RARGC
#define RARGV      OPENISIS_RARGV
#define RFDT       OPENISIS_RFDT
#define RDIS       OPENISIS_RDIS
#define RNOC       OPENISIS_RNOC
#define RIGN       OPENISIS_RIGN
#define RCHG       OPENISIS_RCHG
#define RDEL       OPENISIS_RDEL
#define ROP        OPENISIS_ROP

#define Set        OpenIsisSet
#define Val        OpenIsisVal
#define Key        OpenIsisKey
#define Hit        OpenIsisHit
#define XKEY       OPENISIS_XKEY
#define XVAL       OPENISIS_XVAL
#define XHIT       OPENISIS_XHIT
#define XHEX       OPENISIS_XHEX
#define XDOT       OPENISIS_XDOT
#define XFST       OPENISIS_XFST
#define dRead      openIsisDRead
#define dRaw       openIsisDRaw
#define dFmt       openIsisDFmt
#define dMaxId     openIsisDMaxId
#define dWrite     openIsisDWrite
#define dWritex    openIsisDWritex
#define dScan      openIsisDScan
#define dTerm      openIsisDTerm
#define dQuery     openIsisDQuery
#define DXCb       OpenIsisDXCb
#define DXLoop     OpenIsisDXLoop
#define IDXEQ      OPENISIS_IDXEQ
#define IDXPF      OPENISIS_IDXPF
#define IDXUPTO    OPENISIS_IDXUPTO
#define IDXINCL    OPENISIS_IDXINCL
#define IDXMODE    OPENISIS_IDXMODE
#define dXLoop     openIsisDXLoop

#define Idx        OpenIsisIdx
#define Index      OpenIsisIndex /* Idx* */
#define CLockFunc  OpenIsisCLockFunc
#define cOpen      openIsisCOpen
#define cInit      openIsisCInit
#define cLog       openIsisCLog
#define cSession   openIsisCSession
#define cDOpen     openIsisCDOpen
#define cDOpenv    openIsisCDOpenv
#define cDCheck    openIsisCDCheck
#define cDClose    openIsisCDClose
#define cXOpen     openIsisCXOpen
#define cXMkVal    openIsisCXMkVal
#define cXAdd      openIsisCXAdd
#define cXClose    openIsisCXClose
#define cFdt       openIsisCFdt
#define cFreeFdt   openIsisCFreeFdt

#define stub0      openisis_stub0
#define ses0       openisis_ses0

extern Session ses0;

/*
	the following are not yet blessed by an official OpenIsis prefix
*/
/* lio */
#define Buf        LioBuf
#define ioStream   lio_stream
#define ioStdio    lio_stdio

/* lsv */
#define Que        LsvQue
#define Pool       LsvPool
#define Wrk        LsvWrk
#define Srv        LsvSrv
#define Con        LsvCon
#define svCur      lsv_cur
#define svPlain    lsv_plain
#define svEcho     lsv_echo
#define svRun      lsv_run

#define LOI_H
#endif /* LOI_H */
