/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2001-2002 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */

package org.openisis;

import java.io.*;
import java.util.Arrays;

/**
	Charset tests and utils for openisis java binding.
<p>
	$Id: Charset.java,v 1.2 2002/12/02 13:46:34 kripke Exp $
	@version $Revision: 1.2 $
	@author  $Author: kripke $
*/
public class Charset {

	static final String[] encodings = {
		"ISO-8859-5", // cyrillic
		"ISO-8859-7", // greek
		"ISO-8859-9", // turkish
	};

	// bytes 160(=-96) to 255(=-1)
	static final byte[] isohigh = new byte[96];
	static {
		for ( int b=0; b<96; b++ )
			isohigh[b] = (byte)(b-96);
	}

	static void utf8seq ()
		throws IOException
	{
		Writer w = new OutputStreamWriter( System.out, "UTF-8" );
		for ( int i=0; i<encodings.length; i++ ) {
			String e = encodings[i];
			w.write( e+"|"+(new String(isohigh,e))+"\n" );
		}
		w.flush();
	}


	static void chartest ()
		throws IOException
	{
		String[] args = new String[] { "-v", "error", "-encoding", "UTF-8",
			"-format", "aligned" };
		Db db = Db.open( "db/charset/charset", args );
		for ( int i=0; i<encodings.length; i++ ) {
			String e = encodings[i];
			int[] n = db.search( e, 0 );
			if ( null == n || 1 != n.length ) {
				System.err.println( "bad res "+n+" for encoding "+e );
				continue;
			}
			Rec r = db.readRow( n[0] );
			if ( null == r ) {
				System.err.println( "could not read mfn "+n[0]+" for encoding "+e );
				continue;
			}
			if ( 2 > r.getLen() || null == r.getField(0) || null == r.getField(1) ) {
				System.err.println( "bad fields on mfn "+n[0]+" for encoding "+e );
				continue;
			}
			Field f = r.getField(0);
			if ( 1 != f.tag || !e.equals( f.val ) ) {
				System.err.println( "bad field 0 "+f.tag+"="+f.val
					+" on mfn "+n[0]+" for encoding "+e );
				continue;
			}
			f = r.getField(1);
			if ( 2 != f.tag ) {
				System.err.println( "bad field 1 "+f.tag+"="+f.val
					+" on mfn "+n[0]+" for encoding "+e );
				continue;
			}
			if ( ! f.val.equals( new String(isohigh,e) ) ) {
				System.err.println( "UTF-8 mismatch on mfn "+n[0]+" for encoding "+e );
				continue;
			}
			byte[] b = f.val.getBytes(e);
			if ( ! Arrays.equals( b, isohigh ) ) // ok for some encodings
				System.err.println( "note: no round-trip for encoding "+e );
		}
	}


	public final String enc;

	public Charset ( String encoding ) { enc = encoding; }

	public OutputStreamWriter osw ( OutputStream s )
		throws IOException
	{
		return new OutputStreamWriter( s, enc );
	}

	public void chartab ( OutputStreamWriter w )
		throws IOException
	{
		if ( null == w )
			w = new OutputStreamWriter( System.out );
		w.write( "... 0 1 2 3 4 5 6 7 8 9 A B C D E F\n" );
		byte[] b = new byte[32];
		for ( int l=0; l<8; l++ ) {
			for ( int j=0; j<16; j++ ) {
				b[2*j] = (byte)' ';
				b[2*j+1] = (byte)(l*16 + j - 128);
			}
			String s = new String( b, enc );
			w.write( (128+l*16) + s + "\n" );
			// char[] c = s.toCharArray[];
		}
		w.flush();
	}

	public void codetab ( OutputStreamWriter w )
		throws IOException
	{
		if ( null == w )
			w = new OutputStreamWriter( System.out );
		byte[] b = new byte[16];
		for ( int l=0; l<8; l++ ) {
			for ( int j=0; j<16; j++ )
				b[j] = (byte)(l*16 + j - 128);
			String s = new String( b, enc );
			char[] c = s.toCharArray();
			for ( int j=0; j<c.length; j++ )
				w.write( " " + (int)c[j] + "," );
			w.write( "\n" );
		}
		w.flush();
	}

	public void htmltab ( OutputStreamWriter w )
		throws IOException
	{
		if ( null == w )
			w = osw( System.out );
		w.write( "<html><head>\n"
		+ "<meta http-equiv=\"Content-Type\""
		+ " content=\"text/html;charset="+enc+"\"/>\n"
		+ "</head><body><pre>\n"
		);
		byte[] b = new byte[1];
		for ( b[0] = -96; b[0]<-1; b[0]++ ) {
			w.write( (256+(int)b[0]) + " = "
				+ new String( b, enc ) + "\n" );
		}
		w.write( "</pre></body></html>" );
		w.flush();
	}

	public static void main ( String[] args )
		throws IOException
	{
		String enc = "Cp850";
		char task = 's';
		int a = 0;

		while ( a<args.length ) {
			String arg = args[a++];
			if ( "-enc".equals( arg ) )
				enc = args[a++];
			else if ( "-utf8seq".equals( arg ) )
				task = 'u';
			else if ( "-tab".equals( arg ) )
				task = 't';
			else if ( "-htab".equals( arg ) )
				task = 'h';
			else if ( "-ctab".equals( arg ) )
				task = 'c';
		}

		switch ( task ) {
		case 'u': // -utf8seq
			utf8seq();
			break;
		case 't': // -tab
			(new Charset(enc)).chartab(null);
			break;
		case 'h': // -htab
			(new Charset(enc)).htmltab(null);
			break;
		case 'c': // -ctab
			(new Charset(enc)).codetab(null);
			break;
		case 's': // -test
			chartest();
		}
	}
}	// Charset
