/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2001-2002 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */

package org.openisis;

import java.io.*;
import java.util.*;

import javax.activation.*;
import javax.mail.*;
import javax.mail.internet.*;


/**
	This class contains a collection of javax.mail based E-Mail utilities.

<p>
	$Id: Mail.java,v 1.3 2002/12/04 01:23:47 kripke Exp $
	@version $Revision: 1.3 $
	@author  $Author: kripke $
*/
public abstract class Mail {

	public static final String EMPTY      = "";

	public static final String MESSAGEID  = "message-id";
	public static final String DATE       = "date";
	public static final String TO         = "to";
	public static final String CC         = "cc";
	public static final String BCC        = "bcc";
	public static final String FROM       = "from";
	public static final String SENDER     = "sender";
	public static final String REPLYTO    = "reply-to";
	public static final String SUBJECT    = "subject";
	public static final String REFERENCES = "references";
	public static final String INREPLYTO  = "in-reply-to";
	public static final String KEYWORDS   = "keywords";
	public static final String COMMENTS   = "comments";
	public static final String ENCRYPTED  = "encrypted";
	public static final String PRECEDENCE = "precedence"; // non-standard
	public static final String XPRIORITY  = "x-priority"; // non-standard
	public static final String RETURNPATH = "return-path";
	public static final String DLVRDTO    = "delivered-to"; // qmail
	public static final String RECEIVED   = "received";
	public static final String XMAILER    = "x-mailer";
	public static final String CTYPE      = "content-type";
	public static final String CTRANSENC  = "content-transfer-encoding";

	public static final String TEXT_PLAIN = "text/plain";
	public static final String TEXT_ISO   = "text/plain; charset=\"iso-8859-1\"";
	public static final String TEXT_HTML  = "text/html";
	public static final String MESS_MIME  = "message/rfc822";
	public static final String MULT_MIXED = "multipart/mixed";
	public static final String MULT_ALT   = "multipart/alternative";
	public static final String MULT_APPLE = "multipart/appledouble";
	public static final String MULT_SIGN  = "multipart/signed";
	public static final String MULT_ANY   = "multipart/*";
	public static final String APPL_WORD  = "application/msword";
	public static final String APPL_PDF   = "application/pdf";
	public static final String APPL_RTF   = "application/rtf";
	public static final String APPL_BYTES = "application/octet-stream";
	public static final String APPL_APPLE = "application/applefile";
	public static final String APPL_ANY   = "application/*";
	public static final String IMAG_GIF   = "image/gif";
	public static final String IMAG_JPEG  = "image/jpeg";
	public static final String IMAG_ANY   = "image/*";

	public static final String SMTP_HOST  = "mail.smtp.host";

	public static final Session SES;
	static {
		if ( null == System.getProperty( SMTP_HOST ) )
			System.setProperty( SMTP_HOST, "localhost" );
		SES = Session.getDefaultInstance( System.getProperties(), null );
	}


	public static interface Source {
		Message[] get () throws Exception;
		void done ( boolean expunge ) throws Exception;
	}

	public static interface Sink {
		void put ( Message m ) throws Exception;
		void put ( Message[] m ) throws Exception;
	}


	//	//////////////////////////////////////////////////
	//	utilities
	//

	public static void del ( Message m )
		throws Exception
	{
		m.setFlag( Flags.Flag.DELETED, true );
	}	// del

	public static void del ( Message[] m )
		throws Exception
	{
		for ( int i=0; i<m.length; i++ )
			if ( null != m[i] )
				m[i].setFlag( Flags.Flag.DELETED, true );
	}	// del


	public static void put ( Sink s, Message[] m )
		throws Exception
	{
		for ( int i=0; i<m.length; i++ )
			if ( null != m[i] )
				s.put( m[i] );
	}	// put

	public static int move ( Sink si, Source so )
		throws Exception
	{
		Message[] m;
		int copied = 0;
		while ( null != (m = so.get()) && 0 < m.length ) {
			si.put( m );
			del( m );
			so.done( true );
			copied += m.length;
		}
		return copied;
	}	// move

	public static String head ( Part p, String name, String def )
		throws Exception
	{
		String[] hh = p.getHeader( name );
		return null == hh || 0 == hh.length ? def : hh[0];
	}	// head

	public static String head ( Part p, String name )
		throws Exception
	{
		return head( p, name, null );
	}	// head
	
	// join lines, commonly used for content type
	public static String oneline ( String ct )
	{
		return null == ct ? null : ct.replace( '\n', '\t' ).replace( '\r', ' ' );
	}	// oneline

	public static Message message ( String[] headers, String body )
		throws Exception
	{
		MimeMessage mm = new MimeMessage( SES );
		for ( int i=0; i<headers.length; i+=2 )
			mm.addHeader( headers[i], headers[i+1] );
		mm.addHeader( CTYPE, TEXT_ISO );
		mm.setContent( body, TEXT_ISO );
		if ( null == mm.getSentDate() )
			mm.setSentDate( new Date() );
		return mm;
	}

	//	//////////////////////////////////////////////////
	//	implementations
	//

	public static class Pop implements Source {
		static final FetchProfile FPEMPTY = new FetchProfile();
		Store _store;
		Folder _f;
		String _h;
		int    _o;
		String _u;
		String _p;
		int    _max;

		public Pop ( String host, int port, String user, String pass, int max )
			throws Exception
		{
			_store = SES.getStore( "pop3" );
			_h = host;
			_o = port;
			_u = user;
			_p = pass;
			_max = max;
		}

		public Pop ( String host, String user, String pass, int max )
			throws Exception
		{
			this( host, 110, user, pass, max );
		}

		public Pop ( String host, String user, String pass )
			throws Exception
		{
			this( host, user, pass, 0 );
		}

		public Message[] get ()
			throws Exception
		{
			_store.connect( _h, _o, _u, _p );
			_f = _store.getFolder( "INBOX" ); // the magic default folder
			_f.open( Folder.READ_WRITE );
			Message[] m = _f.getMessages();
			System.err.println( "pop "+_u+"@"+_h+":"+_o+" new "
				+ (null==m ? -1 : m.length) );
			if ( null == m )
				return null;
			if ( 0 != _max && _max < m.length ) {
				Message[] mm = new Message[_max];
				System.arraycopy( m, 0, mm, 0, _max );
				m = mm;
			}
			_f.fetch( m, FPEMPTY );
			System.err.println( "pop "+_u+"@"+_h+":"+_o+" got "+m.length );
			return m;
		}

		public void done ( boolean expunge )
			throws Exception
		{
			if ( null != _f ) {
				_f.close( expunge );
				_f = null;
			}
			_store.close();
		}
	}	// class Pop


	public static class Dir implements Source, Sink { // the Maildir
		final String _t;
		final String _n;
		final String _c;
		final File _ndir;
		int _lim;

		Msg[] got;

		class Msg extends MimeMessage {
			final File _f;
			Msg ( File f )
				throws Exception
			{
				super( SES, new FileInputStream( f ) );
				_f = f;
				System.err.println( "file "+f.getName()
					+": "+head( this, MESSAGEID, "<?>" ) );
			}
		}

		public Dir ( String base ) 
			throws Exception
		{
			_t = base + File.separator + "tmp";
			_n = base + File.separator + "new";
			_c = base + File.separator + "cur";
			_ndir = new File(_n);
			File t = new File(_t);
			File c = new File(_c);
			if ( ! t.isDirectory() && ! t.mkdirs() )
				throw new FileNotFoundException( _t );
			if ( ! _ndir.isDirectory() && ! _ndir.mkdirs() )
				throw new FileNotFoundException( _n );
			if ( ! c.isDirectory() && ! c.mkdirs() )
				throw new FileNotFoundException( _c );
		}

		public String toString () {
			return "Dir "+_ndir.getParentFile().getAbsolutePath();
		}


		public Message[] get () throws Exception
		{
			File[] l = _ndir.listFiles();
			System.err.println( _ndir+": "+l.length );
			int i = l.length;
			if ( 0 != _lim && i > _lim )
				i = _lim;
			got = new Msg[i];
			while ( 0 < i-- ) {
				File f = new File( _t+File.separator+l[i].getName() );
				l[i].renameTo( f );
				got[i] = new Msg( f );
			}
			return got;
		}

		public void done ( boolean expunge ) throws Exception
		{
			if ( expunge && null != got )
				for ( int i=0; i<got.length; i++ )
					if ( got[i].isSet( Flags.Flag.DELETED ) )
						got[i]._f.delete();
			got = null;
		}


		public void put ( Message m )
			throws Exception
		{
			File f;
			String s = File.separator+System.currentTimeMillis();
			while ( ! (f = new File( _t + s )).createNewFile() )
				s += (int)(10*Math.random()) % 10;
			OutputStream o = new FileOutputStream(f);
			m.writeTo( o );
			o.close();
			while ( ! f.renameTo(new File( _n + s )) )
				s += (int)(10*Math.random()) % 10;
		}

		public void put ( Message[] m )
			throws Exception
		{
			Mail.put( this, m );
		}
	}	// class Dir


	public static class Arc implements Sink { // the archive
		static final String OK = "ok";
		static final Map OKHDR = new HashMap();
		static {
			OKHDR.put( MESSAGEID, OK );
			OKHDR.put( DATE, OK );
			OKHDR.put( FROM, OK );
			OKHDR.put( REPLYTO, OK );
			OKHDR.put( SUBJECT, OK );
			OKHDR.put( REFERENCES, OK );
			OKHDR.put( INREPLYTO, OK );
			OKHDR.put( KEYWORDS, OK );
			OKHDR.put( COMMENTS, OK );
		}

		final String _u;
		final String _b;
		final Sink   _fwd;

		final int _maxlines = 60;
		final int _minlines = 15;
		final boolean _zip  = true;


		/** path is /m/d/h/xxx.
		january is Calendar.JANUARY, which is 0
		*/
		public static String path ( int mon, int day, int hour, int min, int sec )
		{
			char mo = (char)('a'+mon);
			char d = (char)(day + (day<10 ? '0' : 'a'-10));
			char h = (char)('a'+hour);
			int hs =  60*min + sec;
			sec = hs - 100*(min = hs / 100);
			char m = (char)(min + (min<10 ? '0' : 'a'-10));
			return File.separator+mo + File.separator+d +File.separator+h
				+	File.separator+m+(char)('0'+sec/10)+(char)('0'+sec%10);
		} // path

		public static String path ( Calendar c )
		{
			return path( c.get( Calendar.MONTH ), c.get( Calendar.DATE ),
				c.get( Calendar.HOUR_OF_DAY ), c.get( Calendar.MINUTE ),
				c.get( Calendar.SECOND ) );
		} // path

		public static String path ( Date d )
		{
			Calendar c = new GregorianCalendar();
			c.setTime( d );
			return path( c );
		}	// path

		public static String path ()
		{
			return path( new GregorianCalendar() );
		}	// path


		public Arc ( String base, String url, Sink fwd ) 
			throws Exception
		{
			_b = base;
			_u = url;
			_fwd = fwd;
			File b = new File(_b);
			if ( ! b.isDirectory() && ! b.mkdirs() )
				throw new FileNotFoundException( _b );
		}

		public String toString () {
			return "Arc "+_b+" "+_u;
		}

		public void put ( Message m )
			throws Exception
		{
			String mid = head( m, MESSAGEID, "<?>" );
			System.err.println( _b+" += "+mid );
			Date d = m.getSentDate();
			if ( null == d )
				d = m.getReceivedDate();
			if ( null == d ) {
				System.err.println( "warning: no date found in "+mid );
				d = new Date();
			}
			String base = path( d );
			String path = base;
			File f = new File( _b + path );
			File dir = f.getParentFile();
			if ( ! dir.isDirectory() && ! dir.mkdirs() )
				throw new FileNotFoundException( dir.getAbsolutePath() );
			for ( int i=0; ! f.createNewFile(); i++ )
				f = new File( _b + (path = base + Integer.toString( i, 36 )) );
			String name = f.getName();
			// got unique path path
			System.out.println( "new message "+mid+" date "+d+" -> "+path );
			MimeMessage mm = new MimeMessage( SES );
			Enumeration e = m.getAllHeaders();
			while ( e.hasMoreElements() ) {
				Header h = (Header)e.nextElement();
				String n = h.getName().toLowerCase();
				if ( OK == OKHDR.get(n) )
					mm.addHeader( n, h.getValue() );
			}
			mm.addHeader( CTYPE, TEXT_ISO );
			mm.addHeader( CTRANSENC, "8bit" );
			StringBuffer b = new StringBuffer( 4096 );
			Stack s = new Stack();
			s.push( m );
			int i = 0;
			int totlines = 0;
			List binparts = new ArrayList( 32 );
			while ( ! s.isEmpty() ) {
				Part p = (Part)s.pop();
				String ct = oneline( p.getContentType() );
				// ct = null == ct ? TEXT_PLAIN : ct.toLowerCase();

				if ( p.isMimeType( MULT_ANY ) ) {
					Object content = p.getContent();
					if ( ! (content instanceof Multipart) ) {
						// ignore
						System.err.println( "multipart type "+ct
							+" has class "+content.getClass().getName() );
						mm.addHeader( COMMENTS, "dropped bad multipart "+ct );
					} else {
						Multipart mu = (Multipart)content;
						int cnt = mu.getCount();
						if ( cnt > 8 ) {
							System.err.println( "multipart with too many parts: "+cnt );
							cnt = 8;
						}
						Part[] pu = new Part[cnt];
						for ( int j=0; j<cnt; j++ )
							pu[j] = mu.getBodyPart( j );
						if ( p.isMimeType( MULT_ALT )
							|| p.isMimeType( MULT_SIGN )
						) { // alternatives suggested
							int select = -1;
							for ( int j=0; j<cnt; j++ ) // see wether we can decide
								if ( pu[j].isMimeType( TEXT_PLAIN ) ) {
									select = j;
									break;
								}
							if ( 0 <= select ) {
								for ( int j=0; j<cnt; j++ )
									if ( select != j )
										mm.addHeader( COMMENTS, "dropped alternative "+
											oneline( pu[j].getContentType() ) + " for "+i );
								if ( 0 != select )
									pu[0] = pu[select];
								cnt = 1;
							}
						} else if ( p.isMimeType( MULT_APPLE ) // apple ...
							&& 2 == cnt // ... double
						) { // kill ressource fork
							int j = 0;
							if ( pu[0].isMimeType( APPL_APPLE )
								|| pu[j=1].isMimeType( APPL_APPLE )
							) {
								if ( 0 == j )
									pu[0] = pu[1];
								mm.addHeader( COMMENTS, "dropped appledouble" );
								cnt = 1;
							}
						}
						for ( int j=cnt; j-- > 0; ) // push parts backwards
							s.push( pu[j] );
					}
					continue;
				} // multi

				i++; // about to create part i
				String part = path+'-'+i;

				// if ( ! p.isMimeType( TEXT_PLAIN ) ) {
				/*
				it's so awfully unreliable ...  many RTFs go as APPL_WORD ...
				if ( p.isMimeType( TEXT_HTML ) )
					ext = ".html";
				else if ( p.isMimeType( APPL_WORD ) )
					ext = ".doc";
				else
					ext = "";
				*/
				OutputStream o = new FileOutputStream( _b+part );
				p.getDataHandler().writeTo( o );
				o.close();
				// can we convert it ?
				String[] args = new String[] { "any2txt", "-v", name+'-'+i };
				Process fix = Runtime.getRuntime().exec( args, null, dir );
				BufferedReader err = new BufferedReader( new InputStreamReader(
					fix.getErrorStream() ) );
				String line;
				while ( null != (line = err.readLine()) )
					System.err.println( "any2txt:\t"+line );
				err.close();
				fix.waitFor();
				int ret = fix.exitValue();
				String ext = EMPTY;
				boolean havetext = 0 != (ret & 1) || 0 == ret;
				int type = ret & ~1;
				switch ( type ) {
				case 0: ext = ".txt"; break;
				case 2: ext = ".html"; break;
				case 4: ext = ".pdf"; break;
				case 6: ext = ".rtf"; break;
				case 8: ext = ".doc"; break;
				case 16: // mail -- reread and start over
					try {
						s.push( new MimeMessage( SES, new FileInputStream( _b+part ) ) );
						// TODO: delete
						continue;
					} catch (Exception ee) {
						System.err.println( "could not reload mail "+mid+"["+i+"] "+ct
							+" from "+part );
						ee.printStackTrace();
					}
					ext = ".eml";
					break;
				case 32: ext = ".gif"; break;
				case 34: ext = ".jpg"; break;
				case 36: ext = ".bmp"; break;
				case 38: ext = ".png"; break;
				case 40: ext = ".tiff"; break;
				}
				System.err.println( mid+"["+i+"] "+ct+" is "+ext+" ("+ret+")" );

				if ( EMPTY != ext && 1 != ret ) // rename
					(new File(_b+part)).renameTo( new File(_b+part+ext) );

				if ( 0 != type ) {
					binparts.add( name+'-'+i+ext );
					// mm.addHeader( COMMENTS, "part "+i+" "+ct+" is "+part );
					if ( ! havetext ) {
						if ( ! _zip ) {
							if ( 0 < totlines )
								b.append( "\r\n---\r\n\r\n" );
							b.append( "\r\nget part "+i+" at "+_u+part+ext+"\r\n" );
						}
						totlines++;
						continue;
					}
				}
				// }

				// now _b+part.txt contains plaintext
				File plain = new File(_b+part+".txt");
				BufferedReader br = new BufferedReader(
					new InputStreamReader( new FileInputStream( plain ) ) );
				int lines = 0;
				int max = _maxlines - totlines;
				if ( max < _minlines )
					max = _minlines;
				// String line;
				if ( 0 < totlines )
					b.append( "\r\n---\r\n\r\n" );
				while (  null != (line = br.readLine()) ) {
					b.append( line ).append( "\r\n" );
					if ( max == ++lines ) {
						mm.addHeader( COMMENTS, "plaintext "+i+" continued at "+part );
						b.append( "... read more: "+_u+part+".txt ("
							+(1 + plain.length()/1024)+" KB)\r\n" );
						break;
					}
				}
				if ( 0 != type && !_zip )
					b.append( "\r\nget original part "+i+" at "+_u+part+ext+"\r\n" );
				totlines += lines;
				br.close();
			}
			int nparts = binparts.size();
			if ( _zip && 0 != nparts ) {
				String[] args = new String[3+nparts];
				int a=0;
				args[a++] = "zip";
				args[a++] = "-m"; // move
				args[a++] = name+".zip";
				for ( int p = 0; p<nparts; p++ )
					args[a++] = (String)binparts.get(p);
				Process zip = Runtime.getRuntime().exec(
					args, null, dir );
				BufferedReader out = new BufferedReader( new InputStreamReader(
					zip.getInputStream() ) );
				String line;
				while ( null != (line = out.readLine()) )
					b.append( line+"\r\n" );
					// System.err.println( "zip:\t"+line );
				out.close();
				zip.waitFor();
				int ret = zip.exitValue();
				File zipped = new File(_b+path+".zip");
				if ( 0 != ret )
					System.err.println( "zip exited "+ret );
				else if ( ! zipped.exists() )
					System.err.println( "no zipfile "+zipped );
				else
					b.append( "\r\nget original parts at "+_u+path+".zip ("
						+(1 + zipped.length()/1024)+" KB)\r\n" );
			}
			mm.setContent( b.toString(), TEXT_ISO );
			OutputStream o = new FileOutputStream(f);
			mm.writeTo( o );
			o.close();
			f.renameTo( new File( f.getPath()+".mail" ) );
			if ( null != _fwd )
				_fwd.put( mm );
		}

		public void put ( Message[] m )
			throws Exception
		{
			Mail.put( this, m );
		}
	}	// class Arc


	public static class Smtp implements Sink {

		Transport _t;
		Address[] _a;
		boolean   _conn;

		public Smtp ( Session s, Address[] a ) throws Exception {
			_t = s.getTransport( s.getProvider("smtp") );
			_a = a;
		}

		public Smtp ( Address[] a ) throws Exception {
			this( SES, a );
		}

		public Smtp ( String a ) throws Exception {
			this( SES, new Address[] { new InternetAddress(a) } );
		}

		/** connect. return true, iff we weren't already connected. */
		public boolean connect () throws Exception {
			if ( _conn )
				return false;
			_t.connect();
			return _conn = true;
		}

		public void close () throws Exception {
			if ( _conn ) {
				_t.close();
				_conn = false;
			}
		}

		public void put ( Message m )
			throws Exception
		{
			boolean nconn = connect();
			_t.sendMessage( m, null != _a ? _a : m.getAllRecipients() );
			if ( nconn ) close();
		}

		public void put ( Message[] m )
			throws Exception
		{
			boolean nconn = connect();
			Mail.put( this, m );
			if ( nconn ) close();
		}
	}	// class Smtp


	public static class Cmd implements Sink {

		String _cmd;

		public Cmd ( String cmd ) {
			_cmd = cmd;
		}

		public void put ( Message m )
			throws Exception
		{
			Process run = Runtime.getRuntime().exec( _cmd );
			OutputStream o = run.getOutputStream();
			m.writeTo( o );
			o.close();
			BufferedReader err = new BufferedReader( new InputStreamReader(
				run.getErrorStream() ) );
			String line;
			while ( null != (line = err.readLine()) )
				System.err.println( _cmd+":\t"+line );
			err.close();
			run.waitFor();
			int ret = run.exitValue();
			if ( 0 != ret )
				System.err.println( "Cmd '"+_cmd+"' exited "+ret );
		}

		public void put ( Message[] m )
			throws Exception
		{
			Mail.put( this, m );
		}
	}	// class Cmd


	public static void main ( String[] args )
		throws Exception
	{
		String host = "localhost";
		int    port = 110;
		String user = "test";
		String pass = "test";
		String sodir = "/tmp/test";
		String sidir = "/tmp/test";
		String url  = "http://openisis.org/archive";
		int a = 0;
		Source so = null;
		Sink si = null;
		Sink fwd = null;
		char source = 'd';
		char sink = 'd';
		char task = 0;
		int n = 0;

		while ( a<args.length ) {
			String arg = args[a++];
			if ( "-h".equals( arg ) )
				host = args[a++];
			else if ( "-o".equals( arg ) )
				port = Integer.parseInt( args[a++] );
			else if ( "-u".equals( arg ) )
				user = args[a++];
			else if ( "-p".equals( arg ) )
				pass = args[a++];
			else if ( "-pop".equals( arg ) )
				source = 'p';
			else if ( "-url".equals( arg ) ) // arc
				url = args[a++];
			else if ( "-fwd".equals( arg ) ) // fwd by smtp
				fwd = new Smtp( args[a++] );
			else if ( "-fwdcmd".equals( arg ) ) // fwd by command
				fwd = new Cmd( args[a++] );
			else if ( "-dir".equals( arg ) ) {
				source = 'd';
				sodir = args[a++];
			} else if ( "-todir".equals( arg ) ) {
				sink = 'd';
				sidir = args[a++];
			} else if ( "-toarc".equals( arg ) ) {
				sink = 'a';
				sidir = args[a++];
			} else if ( "-move".equals( arg ) )
				task = 'v';
			else
				throw new IllegalArgumentException( arg );
		}

		switch ( source ) {
		case 'p':
			so = new Pop( host, port, user, pass, 8 );
			break;
		default: // 'd'
			so = new Dir( sodir );
		}
		switch ( sink ) {
		case 'a':
			si = new Arc( sidir, url, fwd );
			break;
		default: // 'd'
			si = new Dir( sidir );
		}

		switch ( task ) {
		case 'v': // -move
			n = move( si, so );
			System.err.println( "moved "+n+" messages from "+so+" to "+si );
			break;
		}
	}
}	// Mail
