/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2001-2002 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */

package org.openisis;

import java.io.IOException;
import java.util.Date;
import java.util.Iterator;
import java.text.DecimalFormat;


/**
	Test of openisis java binding.
<p>
	$Id: Test.java,v 1.3 2003/03/05 19:52:29 istr Exp $
	@version $Revision: 1.3 $
	@author  $Author: istr $
*/
class Test extends Thread {

	static DecimalFormat mfn4 = new DecimalFormat( "0000" );

	public static boolean dumpRow ( Db db, Rec r, int m ) {
		if ( null == r ) {
			System.err.println( "could not get row " );
			return false;
		}
		if ( 1 == m ) { // cds1 test special
			m = Db.MDL | Db.PS;
			String r13 = "\n             ", s = "\n";
			System.out.println( "MFN        : "+mfn4.format(r.getRowId()) );
			System.out.print( r.v("CONFERENCE : ",null,12,m,0,0,r13,s) );
			System.out.print( r.v("TITLE      : ",null,24,m,0,0,r13,s) );
			System.out.print( r.v("AUTHOR(S)  : ",null,14,m,0,0,"; "+r13+" ",s) );
			System.out.print( r.v("EDITION    : ",null,25,m,0,0,r13,s) );
			System.out.print( r.v("IMPRINT    : ",null,26,m,0,0,r13,s) );
			System.out.print( r.v("COLLATION  : ",null,30,m,0,0,r13,s) );
			System.out.print( r.v("SERIES     : ",null,44,m,0,0,r13,s) );
			System.out.print( r.v("NOTES      : ",null,50,m,0,0,r13,s) );
			System.out.print( r.v("CORP. BODY : ",null,71,m,0,0,r13,s) );
			System.out.print( r.v("ADDED CONF : ",null,72,m,0,0,r13,s) );
			System.out.print( r.v("OTHER TITLE: ",null,74,m,0,0,r13,s) );
			System.out.print( r.v("OTHER LANG : ",null,76,m|'*',0,0,r13,s) );
			System.out.print( r.v("KEYWORDS   : ",null,69,m,0,0,r13,s) );
			System.out.print( "\n\n" );
			return true;
		}
		System.out.println( "got row "+r.getRowId() );
		int l = r.getLen();
		for ( int i=0; i<l; i++ ) {
			Field f = r.getField(i);
			System.out.println( "<"+f.tag+">"+f.v(m)+"</"+f.tag+">" );
			if ( 0 == (Db.MHL & m) && '^' == f.val.charAt(0) ) {
				Field[] s = f.split();
				if ( null == s )
					continue;
				for ( int j=0; j<s.length; j++ )
					if ( null != s[j] )
						System.out.println( "<"+f.tag+(char)s[j].tag+">"
							+s[j].v(m) +"</"+f.tag+(char)s[j].tag+">" );
			}
		}
		return true;
	}	// dumpRow

	public static boolean dumpRow ( Db db, Rec r ) {
		return dumpRow( db, r, Db.HTU );
	}	// dumpRow


	static String[] term, val;
	static Db db;
	static int[][] post;

	int start;

	Test ( int s ) {
		start = s;
	}

	/** implementation of runnable for crashtest.  */
	public void run () {
		long then = (new Date()).getTime();
		for ( int i=start+1; i++!=start; ) { // read recs for term i
			if ( i >= term.length )
				i = 0;
			// System.err.println( "post "+i );
			for ( int j=0; j<post[i].length; j++ ) {
				int mfn = post[i][j];
				Rec r;
				if ( null == (r = db.readRow( mfn )) )
					System.err.println( "no rec "+mfn );
				else if ( ! val[mfn].equals( r.getField( 0 ).val ) )
					System.err.println( "mismatch on "+mfn );
			}
		}
		long now = (new Date()).getTime();
		System.out.println( "thread " + Thread.currentThread()
			+ " terminated after " + ((now - then)/1000.) + " seconds" );
	}	// run

	public static boolean crash ( String pre )
		throws InterruptedException
	{
		// collect terms array
		int n = 0;
		for ( Iterator i=db.terms(pre); i.hasNext(); n++ )
			i.next();
		System.out.println( n + " terms" );
		term = new String[n];
		post = new int[n][];
		n = 0;
		int p = 0;
		int max = 0;
		for ( Iterator i=db.terms(pre); i.hasNext(); n++ ) {
			post[n] = db.search( term[n] = (String) i.next(), Db.QRY_KEYEQ );
			p += post[n].length;
			for ( int j=0; j<post[n].length; j++ )
				if ( max < post[n][j] )
					max = post[n][j];
		}
		System.out.println( p + " postings max mfn " + max );
		val = new String[max+1];
		Rec r;
		Field f;
		long then = (new Date()).getTime();
		for ( int i=1; i<val.length; i++ )
			if ( null == (r = db.readRow( i )) )
				System.err.println( "no rec "+i );
			else if ( 0 < r.getLen() )
				val[i] = r.getField( 0 ).val;
		long now = (new Date()).getTime();
		System.out.println( "sequential read " + max + " rows in "
			+((now - then)/1000.) + " seconds "
			+(max * 1000 / (now - then)) + " rows per sec"
		);

		int[] nThreads = { 8, 2, 1, 4 };
		int[] res = new int[nThreads.length];

		int passes = 3;

		for ( int pass=0; pass<passes; pass++ )
			for ( int i=0; i<nThreads.length; i++ ) {
				Thread[] th = new Thread[ nThreads[i] ];
				then = (new Date()).getTime();
				for ( int j=0; j<nThreads[i]; j++ )
					(th[j] = new Test( j*term.length/nThreads[i] )).start();
				System.out.println( "started "+nThreads[i]+" threads" );
				for ( int j=0; j<nThreads[i]; j++ )
					th[j].join();
				now = (new Date()).getTime();
				int rps = (int)(nThreads[i] * p * 1000 / (now - then));
				System.out.println( "joined "+nThreads[i]+" threads after "
					+((now - then)/1000.) + " seconds " +rps + " rows per sec"
				);
				res[i] += rps;
			}
		for ( int i=0; i<nThreads.length; i++ )
			System.out.println( nThreads[i]+" threads  "
				+(res[i]/passes) + " rows per sec" );
		return true;
	}	// crash


	public static void main ( String[] args )
		throws IOException, InterruptedException
	{
		int m = Db.HTU;
		int a = 0; // args offset
		db = Db.open( null, args );
		Rec r;
		// System.err.println( "db = "+db );

		while ( a < args.length ) {

			if ( ! args[a].startsWith("-") ) { // a number ?
				try {
					dumpRow( db, db.readRow( Integer.parseInt(args[a]) ), m );
				} catch (NumberFormatException e) {}
				a++;
				continue;
			}

			if ( args[a].startsWith("-m") ) { // -mxx option
				if ( 2 < args[a].length() )
					switch ( args[a].charAt(2) ) {
					case 'h': m = Db.MHL; break;
					case 'd': m = Db.MDL; break;
					case '1': m = 1; /* cds1 impl by dumprow */ break;
					default: m = 0;
					}
				if ( 3 < args[a].length() && 'u' == args[a].charAt(3) )
					m |= Db.MXU;
				a++;
				continue;
			}

			if ( args[a].startsWith("-ht") ) { // -htx option
				m |= Db.HTU;
				if ( 3 < args[a].length() )
					switch ( args[a].charAt(3) ) {
					case 'i': m |= Db.HTI; break;
					case 'a': m |= Db.HTA; break;
					}
				a++;
				continue;
			}

			if ( "-dump".equals(args[a]) ) { // full dump
				for ( int i=1; dumpRow( db, db.readRow(i), m ); i++ )
					;
				a++;
				continue;
			}

			if ( "-write".equals( args[a] ) ) { // read a row, modify and write it, read again and check
				Rec r1 = db.readRow(1);
				dumpRow( db, r1, m );

				Field[] fs = r1.getFields();
			
				fs[ 0 ] = new Field( fs[ 0 ].tag, fs[ 0 ].val + "TEST" );
				
				Rec r2 = new Rec( r1.getRowId(), fs ); 
				
				System.err.println( "write result is: " + db.writeRow( r2 ) );
				r1 = db.readRow(1);
				dumpRow( db, r1, m );
			}

			if ( "-writex".equals( args[a] ) ) { // read row, modify it and write w/ index
				Rec r1 = db.readRow(1);
				dumpRow( db, r1, m );

				Field[] fs = r1.getFields();

				// TODO: impl some test here
			}

			// all other options, known or not, are assumed to need a param
			// i.e. 2 args
			if ( 2+a > args.length )
				break;

			if ( "-terms".equals(args[a]) )
				for ( Iterator i=db.terms(args[a+1]); i.hasNext(); )
					System.out.println( i.next() );
			else if ( "-row".equals(args[a]) )
				dumpRow( db, db.readRow( Integer.parseInt(args[a+1]) ), m );
			else if ( "-scan".equals(args[a]) )
				for ( int i=1; null != (r = db.scanRow(i,-42,args[a+1])); i++ ) {
					dumpRow( db, r, m );
					i = r.getRowId();
				}
			else if ( "-search".equals(args[a]) || "-query".equals(args[a]) ) {
				String src = args[a+1];
				int mode = 0;
				int[] set = db.search(src,
					"-search".equals(args[a]) ?  Db.QRY_KEYAT : Db.QRY_SIMPLE
				);
				System.err.println( set.length + " rows" );
				for ( int i=0; i<set.length; i++ )
					System.err.print( set[i] + "\t" );
				System.err.println( "" );
				for ( int i=0; i<set.length; i++ )
					dumpRow( db, db.readRow( set[i] ), m );
			}
			else if ( "-crash".equals(args[a]) ) { // crash test
				crash( args[a+1] );
			}
			

			a += 2;
		}	// while
	}	// main

}	// Test
