/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */

package org.openisis;

import java.util.*;

/**
	Various utilities.
<p>
	$Id: Utils.java,v 1.1 2003/03/28 08:26:32 mawag Exp $
	@version $Revision: 1.1 $
	@author  $Author: mawag $
*/

public abstract class Utils {

	private Utils () {}

	/**	Build field array from map. Map keys must be valid field description
		names (if fdt is given) or objects whose string representation
		build tag numbers.
		@param ign if false, method will return null on an occurence of a
			map entry that cannot be transformed in a field
		@return field array or null, if ign is false and map contains
			illegal entries
	*/
	public static Field[] ToFields (Map map, Fdt fdt, boolean ign) {
		ArrayList  lst  =  new ArrayList ();
		Iterator   it;
		Fdt.Fd     fd;
		Map.Entry  me;
		String     str;
		Object     key, val;
		int        tag;

		if (null != map) {
			it = map.entrySet ().iterator ();
			while (it.hasNext ()) {
				me = (Map.Entry) it.next ();
				key = me.getKey ();
				if (null == key) {
					if (ign) {
						continue;
					}
					return null;
				}
				if (key instanceof Number) {
					tag = ((Number)key).intValue ();
				}
				else {
					str = key.toString ();
					try {
						tag = Integer.parseInt (str);
					}
					catch (Exception ex) {
						fd = null;
						if (null != fdt) {
							fd = fdt.fdByName (str);
						}
						if (null == fd) {
							if (ign) {
								continue;
							}
							return null;
						}
						tag = fd.id;
					}
				}
				val = me.getValue ();
				str = null == val ? "" : val.toString ();
				lst.add (new Field (tag, str));
			}
		}

		return (Field[]) lst.toArray (new Field [lst.size ()]);
	}

	/**	Fill tgt with field tag-val-entries.
		<br/>
		If fdt is given, only fields
		having a field decription are put in tgt with string keys.
		If fdt is null, entry keys are set as integers.
		<br/>
		If tgt is null, a new HashMap will be allocated.
	*/
	public static Map ToMap (Field[] flds, Fdt fdt, Map tgt) {
		Fdt.Fd fd;

		if (null != flds) {
			if (null == tgt) {
				tgt = new HashMap (flds.length);
			}

			for (int j = flds.length; 0 <= --j;  ) {
				if (null != fdt) {
					if (null != (fd = fdt.fdById (flds[j].tag))) {
						tgt.put (fd.name, flds[j].val);
					}
				}
				else {
					tgt.put (new Integer (flds[j].tag), flds[j].val);
				}
			}
		}

		return tgt;
	}
}

