/*
openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2001-2002 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */


package org.openisis;

/**
	convert Xmldocuments to Record
*/

import org.xml.sax.*;
import org.xml.sax.helpers.*;
import java.io.*;
import java.util.*;


public class XmlToRec {

	private XMLReader myReader;
	private HashMap myTags;
	private HashMap myAttributes;
	private int myMaxTag;
	private int myMaxRowId;
	private OIHandler myHandler;
	
	public XmlToRec(XMLReader reader) {
			myReader=reader;
			myTags=new HashMap();
			myAttributes=new HashMap();
			myMaxTag=0;
			myMaxRowId=0;
			reader.setContentHandler(myHandler=new OIHandler());
	}

	public Fdt getFdt() {
		if (myTags.size()>0) {
			Iterator it=myTags.entrySet().iterator();
			Map.Entry e;
			ArrayList fdl=new ArrayList();
			List att;
			String atts;
			while(it.hasNext()) {
			  	e=(Map.Entry)it.next();
					att=(List)myAttributes.get(e.getKey());
					if (att!=null && att.size()>0) {
						atts="";
						for(int i=0;i<att.size();i++) atts=atts+((char)('a'+i));
					} else atts=null;
					fdl.add(new Fdt.Fd(((Integer)e.getValue()).intValue(),Fdt.FTX,Integer.MAX_VALUE,true,(String)e.getKey(),atts,""));
			}
			Fdt.Fd fds[]=new Fdt.Fd[fdl.size()];
			for(int i=0;i<fds.length;i++) fds[i]=(Fdt.Fd)fdl.get(i);
			Fdt fdt=new Fdt(fds);
			return fdt;
		} else return null;
	}

	public int getTag(List path) {
		StringBuffer tp=new StringBuffer();
		String s;
		Integer ti;
		/*for(int i=0;i<path.size();i++) tp.append(((String)path.get(i)).toLowerCase());
		s=tp.toString();*/
		s=(String)path.get(path.size()-1);
		if ((ti=(Integer)myTags.get(s))!=null) {
			return ti.intValue();
		}
		myMaxTag++;
		myTags.put(s,new Integer(myMaxTag));
		return myMaxTag;
	}

	public char getAttTag(List path,String att) {
		StringBuffer tp=new StringBuffer();
		String s;
		//Integer ti;
		//for(int i=0;i<path.size();i++) tp.append(((String)path.get(i)).toLowerCase());
		s=(String)path.get(path.size()-1);
		ArrayList li;
		if ((li=(ArrayList)myAttributes.get(s))==null) {
			li=new ArrayList();
			myAttributes.put(s,li);
		}
		att=att.toLowerCase();
		int tag;
		tag=li.indexOf(att);
		if (tag<0) {
			tag=li.size();	
			li.add(att);
				
		}
		return (char)(((int)'a')+tag);
	}

	public Rec toRec(Reader reader) {
		try {
		myReader.parse(new InputSource(reader));
		} catch(Exception e) {
			e.printStackTrace();
			return null;
		}
	  return myHandler.getRecord();	
	}

	public Rec toRec(String xml) {
		return toRec(new StringReader(xml));
	}

	public int getNextRowId() {
		return ++myMaxRowId;
	}
	
	
	class OINode {
		String value=null;
		HashMap attributes=null;
		OINode() {
			
		}
	}

	class OIHandler extends DefaultHandler {
	
			private Rec myRec;
			private ArrayList myPath;
			private ArrayList myNodes;
			private ArrayList mySubRec;

			public void startDocument()
                   throws SAXException {
					myPath=new ArrayList();					 
					myNodes=new ArrayList();
					mySubRec=new ArrayList();
					myRec=new Rec(getNextRowId());
			}

			/*public void setRecord(Rec rec) {
				myRec=rec;
			}*/

			public Rec getRecord() {
				return myRec;
			}
		
			public void startElement(java.lang.String uri,
                         java.lang.String localName,
                         java.lang.String qName,
                         Attributes attributes) throws SAXException
			{
				myPath.add(localName);
				OINode node;
				StringBuffer value=new StringBuffer();
				//myNodes.add(node=new OINode());
			  if (attributes.getLength()>0) {
					//node.attributes=new HashMap();
					for(int i=0;i<attributes.getLength();i++) {
					 //node.attributes.put(new Character(getAttTag(myPath,attributes.getLocalName(i))),attributes.getValue(i));
					 value.append('^');
					 value.append(new Character(getAttTag(myPath,attributes.getLocalName(i))));
					 value.append(attributes.getValue(i));
					}
				}
				/*if (mySubRec.size()>0) {
				Object tO=mySubRec.get(mySubRec.size()-1);
				if (tO instanceof Field) {
					mySubRec.remove(mySubRec.size()-1);
					Rec r=new Rec(0);
					r.append((Field)tO);
					mySubRec.add(r);
				}
				}*/
				Rec r=new Rec(0);
				r.append(new Field(getTag(myPath),value.length()==0?null:value.toString()));
				mySubRec.add(r);
				super.startElement(uri,localName,qName,attributes);
			}
			

			public void endElement(java.lang.String uri,
                         java.lang.String localName,
                         java.lang.String qName) throws SAXException {
				if (mySubRec.size()>=2) {
			  Object tO=mySubRec.get(mySubRec.size()-1);
				Rec r=(Rec)mySubRec.get(mySubRec.size()-2);
				if (tO instanceof Field) {
					r.append((Field)tO);
				} else {
					//Rec r2=new Rec(r.getRowId());
					//r2.wrap((Rec)tO,Integer.MAX_VALUE);
					r.wrap((Rec)tO,Integer.MAX_VALUE);
				}
				Rec r2=new Rec(r.getRowId());
				r2.append(r.getFields(0,1));
				r2.wrap(new Rec(0,r.getFields(1,r.getLen())),Integer.MAX_VALUE);
				//r2.wrap(r,Integer.MAX_VALUE);
				mySubRec.set(mySubRec.size()-2,r2);
				} else {
				  if (mySubRec.size()==1) {
					
					myRec=(Rec)mySubRec.get(mySubRec.size()-1);
					//myRec=new Rec(r.getRowId());
					//myRec.wrap(r,Integer.MAX_VALUE);
					//myRec=(Rec)mySubRec.get(mySubRec.size()-1);
					}
				}
				myPath.remove(myPath.size()-1);
				//myNodes.remove(myNodes.size()-1);
				mySubRec.remove(mySubRec.size()-1);
				super.endElement(uri,localName,qName);
			}
			
			public void characters(char[] ch,
                       int start,
                       int length)
                throws SAXException {
        //OINode node=(OINode)myNodes.get(myPath.size()-1);
				Object tO=mySubRec.get(mySubRec.size()-1);
				String value=null;
				Field f=null;
				Rec r=null;
				if (tO instanceof Field) {
					f=(Field)tO;
					value=f.val;
				} else {
				  r=(Rec)tO;
					f=r.get(0);
					value=f.val;
				}
				if (value==null || value.length()==0) {
					value=new String(ch,start,length);
				} else {
					value=new String(ch,start,length)+value;
				}
				if (r==null) {
					mySubRec.set(mySubRec.size()-1,new Field(f.tag,value));
				} else {
					r.set(0,new Field(f.tag,value));
				}
			}

			public void endDocument() throws SAXException {
				//myRec=null;
				super.endDocument();
			}
			
	}
	
}
