#!/bin/bash

#	openisis - an open implementation of the ISIS database
#	Version 0.8.x (microversion see file Version)
#	Copyright (C) 2001-2002 by Erik Grziwotz, erik@openisis.org
#
#	This library is free software; you can redistribute it and/or
#	modify it under the terms of the GNU Lesser General Public
#	License as published by the Free Software Foundation; either
#	version 2.1 of the License, or (at your option) any later version.
#
#	This library is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#	Lesser General Public License for more details.
#
#	You should have received a copy of the GNU Lesser General Public
#	License along with this library; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	This software is dedicated to the memory of Eckart Dietrich.
#
#	This software is inspired by (but contains no code of) the iAPI
#	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
#	See iAPI.txt for what it contains.
#
#	$Id: build-tcltk.sh,v 1.2 2002/12/19 14:43:59 erikisis Exp $
#	build script for tcl/tk

TCL_TK_MAJOR=8.3
TCL_TK_MINOR=4
TCLTK="$TCL_TK_MAJOR.$TCL_TK_MINOR"

not_found () {
	cat <<- eot
		No $1 sources found.
		Please fetch the sources from "http://www.scriptics.com"
		and unpack them in "./$1" or create "./$1" as a symlink 
		to the path, where you have installed the tcl sources.

	eot
	exit 1
}

do_it () {
	while [ 0 -lt $# ]; do
		case $1 in
			--*) extra_args="$extra_args $1";;
			*) dir=$1;;
		esac
		shift
	done
	[ ! -e "${dir}${TCLTK}" ] && not_found "${dir}${TCLTK}"
	( pdir=$(pwd); cd ${dir}${TCLTK}/${tclsubdir};
		grep openisistcl Makefile.in > /dev/null 2>&1 || \
			patch -b -z .orig -p0 < "$pdir/${dir}${TCLTK}${tclsubdir}.patch"
	)
	[ -d ${dir}build ] && rm -rf ${dir}build
	mkdir ${dir}build
	cd ${dir}build
	../${dir}${TCLTK}/$tclsubdir/configure $CONFIGURE_ARGS $extra_args
	make || { echo "error occured"; exit 1; }
	make install
	cd ..
}

#######################################


# change tclsubdir for cross builds
: ${tclsubdir:=unix}

# install dir
PREFIX=$PWD/tcltk/$tclsubdir

build_tcl=false
build_tk=false

while [ 0 -lt $# ]; do
	case $1 in
		-tcl*) build_tcl=true;;
		-tk*)  build_tk=true;;
		*);;
	esac
	shift
done

if ! $build_tk && ! $build_tcl; then
	build_tk=true
	build_tcl=true
fi

CONFIGURE_ARGS="--disable-shared --prefix=$PREFIX"

if [ "win" = "$tclsubdir" ]; then
	PATH=/opt/cross-tools/bin:/opt/cross-tools/i386-mingw32msvc/bin:$PATH
	CONFIGURE_ARGS="$CONFIGURE_ARGS --enable-gcc
		--cache-file=config.cross.cache --host=i386-mingw32msvc 
		--target=i386-mingw32msvc --build=i386-linux"
fi

if $build_tcl; then
	[ -d $PREFIX ] && rm -rf $PREFIX
	do_it tcl 
fi

$build_tk && do_it tk "--with-tcl=../tclbuild"

echo "done"
