/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2001-2002 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */

/*
	$Id: oits.c,v 1.5 2003/03/27 14:49:32 kripke Exp $
	OpenIsis Tcl server
*/


#include "lsv.h"
#include "openisistcl.h"

/*
include this after the Tcl stuff for the benefit of those
who use the 150% braindead gcc 2.96
which barfs on
 declaration of `index' shadows global declaration
in generic/tclDecls.h
cause string.h declares
char *index(const char *s, int c)
*/
#include <string.h> /* mem* */

static Tcl_Interp *interp; /* the 1 + only */
static int oitsid;
static int oitrid[3];


/* the echo application */
static int oits ( Con *c, int task )
{
	Field *f;
	int i;

	switch ( task ) {
	case LSV_APPINI: /* init thread */
	case LSV_APPFIN: /* fini thread */
	case LSV_APPARG: /* new request param arrived */
	default:
		return 0;
	case LSV_APPGOT: /* decide wether request may go to background thread */
		c->grp = 1; /* non Tcl */
		for ( i=c->req->len, f = c->req->field; i--; f++ )
			/* by convention, request with fields >99 are handled by Tcl */
			if ( 99 < f->tag ) {
				c->grp = 0;
				break;
			}
		return 0;
	case LSV_APPRUN:
		break; /* main code below */
	}
	/* run it */
	sMsg( c->ses, LOG_INFO, "echo %d: %d fields to %s",
		svCur()->id, c->req->len, c->nam );
	/**
		we have records:
		req -- the request, allocated by the srv->ses and readonly here
		ses -- session props, allocated and managed by ses, read/write
		res -- the responses, allocated and managed by ses, read/write

		(the identified) ses may be a different session on each call, however,
		calls are strictly serialized (all in main thread)

		alternatively, we may use the one main session bound to the
		main thread as allocator and use the identified session only
		to keep the session props -- given they are never modified
		in non-main-thread calls
	*/
	switch ( c->grp ) {
	case 0: { /* main thread Tcl request */
		Rec *rec[3];
		rec[0] = c->req;
		rec[1] = c->ses->res;
		rec[2] = c->ses->prop;
		if ( openIsisTclEval( oitsid, c->ses, 3, oitrid, rec, "oitscb" ) ) {
			const char *errorInfo
				= Tcl_GetVar( interp, "errorInfo", TCL_GLOBAL_ONLY );
			sMsg( c->ses, LOG_ERROR, "error in callback\n%s",
				errorInfo ? errorInfo : "-" );
		}
		c->req = rec[0];
		c->ses->res = rec[1];
		c->ses->prop = rec[2];
		/* don't break; */
	} /* case 0 */
	default: /* cp session to response */
		if ( c->ses->prop )
			for ( i=c->ses->prop->len, f = c->ses->prop->field; i--; f++ ) {
				RADDF( c->ses, c->ses->res, f, !0 );
			}
		break;
	}
	return 0;
}	/* oits */


#ifdef ORATCL
extern int Oratcl_Init ( Tcl_Interp *ip );
#endif


int main ( int argc, const char **argv )
{
	char *file;
	int ret;
	Srv echo;

	cOpen(0);
	cLog( LOG_DEBUG, 0 );

	interp = Tcl_CreateInterp();
	ret = Tcl_Init( interp );
	sMsg( ses0, LOG_VERBOSE, "appinit %d", ret );
#ifdef ORATCL
	ret = Oratcl_Init( interp );
	sMsg( ses0, LOG_VERBOSE, "orainit %d", ret );
#endif
	oitsid = openIsisTclNewSession( ses0, interp );
	sMsg( ses0, LOG_VERBOSE, "oitsid %d", oitsid );
	oitrid[0] = openIsisTclCreateRecCmd( oitsid, "request", 0 );
	sMsg( ses0, LOG_VERBOSE, "request %d", oitrid[0] );
	oitrid[1] = openIsisTclCreateRecCmd( oitsid, "response", 0 );
	sMsg( ses0, LOG_VERBOSE, "response %d", oitrid[1] );
	oitrid[2] = openIsisTclCreateRecCmd( oitsid, "session", 0 );
	sMsg( ses0, LOG_VERBOSE, "session %d", oitrid[2] );
	file = 2<argc ? (char*)argv[2] : "oitsrc";
	ret = Tcl_EvalFile( interp, file );
	sMsg( ses0, LOG_VERBOSE, "file '%s' %d", file, ret );
	if ( TCL_OK != ret ) {
		const char *errorInfo = Tcl_GetVar( interp, "errorInfo", TCL_GLOBAL_ONLY );
		sMsg( ses0, LOG_ERROR, "error sourcing file '%s' code %d errorInfo\n%s",
			file, ret, errorInfo ? errorInfo : "-" );
		return ret;
	}

	memset( &echo, 0, sizeof(echo) );
	echo.prt = svPlain;
	echo.app = oits;
	return svRun( &echo, 1<argc ? argv[1] : 0 );
}
