#
#/*
#	openisis - an open implementation of the ISIS database
#	Version 0.8.x (microversion see file Version)
#	Copyright (C) 2001-2002 by Erik Grziwotz, erik@openisis.org
#
#	This library is free software; you can redistribute it and/or
#	modify it under the terms of the GNU Lesser General Public
#	License as published by the Free Software Foundation; either
#	version 2.1 of the License, or (at your option) any later version.
#
#	This library is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#	Lesser General Public License for more details.
#
#	You should have received a copy of the GNU Lesser General Public
#	License along with this library; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	This software is dedicated to the memory of Eckart Dietrich.
#
#	This software is inspired by (but contains no code of) the iAPI
#	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
#	See iAPI.txt for what it contains.
#EOH */
#
#
#	$Id: openIsis.tcl,v 1.4 2003/04/07 13:12:23 kripke Exp $
#	the openIsis package
#

package require Tcl 8.3

package provide openIsis 0.8.7

namespace eval openIsis {
	namespace export puts MHL MHLl
}

# Tsp compile
# create Tcl code for Tcl Server Page
# the returned code is usually evaluated in a namespace
proc openIsis::Tspc {tsp {intro ""}} {
	# vars global and main hold global and main code, resp
	set global {}
	set main {}
	# which section we are in
	set sect main
	set i 0
	set lit {$j > $i};# condition for having a literal
	while {"#" == [string index $tsp $i]} {
		set i [string first "\n" $tsp $i]
		if {0 > $i} { error "unterminated line" }
		incr i
	}
	while 1 {
		# we're in literal (outside <% %>) mode
		if {0 <= [set j [string first <% $tsp $i]]} {
			incr j -1
		} else {
			set j end
			set lit {$i <= [string length $tsp]}
		}
		if $lit {
			append $sect "puts " [list [string range $tsp $i $j]] "\n"
		}
		if {"end" == $j} break
		# have <% code, j was last char before
		incr j 3
		if {0 > [set i [string first %> $tsp $j]]} {
			error "unbalanced <% at pos $j '[string range $tsp $j [expr $i+80]]'..."
		}
		incr i -1
		switch -- [set type [string index $tsp $j]] \
		- { # comment -- actually should start --
			# extend to end of comment, skipping all intervening %>
			if {0 > [set i [string first --%> $tsp $j]]} {
				error "unbalanced <%-- at pos $j ..."
			}
			incr i 4 ;# skip --%>
			continue
		} = {
			incr j
			set expr [string range $tsp $j $i]
			# apply some magic guessing
			if [regexp {^[.\-\d]} $expr] { # looks like v get args
				append $sect {puts [lindex [v get } $expr "] 0]\n"
			} elseif [regexp {^[\w\d:]+($|\()} $expr] { # looks like variable
				append $sect "puts \"\$" $expr "\"\n"
			} elseif [regexp {^\w} $expr] { # looks like command
				append $sect {puts [} $expr "]\n"
			} else { # use as is
				append $sect "puts \"" $expr "\"\n"
			}
		} @ {
			error "@-section is meaningless with TSPs"
		} ! {
			set sect global
			if [set globlit [expr {"!" == [string index $tsp $i]}]] {
				incr i -1 
			}
			incr j
		} default {
			set type ""
			set sect main
		}
		if {"=" != $type} {
			append $sect [string range $tsp $j $i] "\n"
		}
		if {"!" == $type} {
			if $globlit {
				incr i 1
			} else { # reset
				set sect main
			}
		}
		incr i 3
	}
	return "$intro
$global
proc main {} {
$main
}
";
}	;# openIsis::Tspc

proc openIsis::puts {str} {
	::puts -nonewline $str
}

# run tspfile, compiling if necessary
# 
proc openIsis::Tspr {tspfile} {
	set compile 1
	set mtime [file mtime $tspfile]
	catch {
		if {[set ::tsp::${tspfile}::mtime] == $mtime} { set compile 0 }
	}
	if $compile {
		catch {namespace delete ::tsp::$tspfile}
		set f [open $tspfile]
		namespace eval ::tsp::$tspfile [Tspc [read $f] "
			variable mtime $mtime
			namespace import ::openIsis::*
		"]
		close $f
		# ::puts stderr "compiled mtime [set ::tsp::${tspfile}::mtime]
		# [info body ::tsp::${tspfile}::main]"
	}
	::tsp::${tspfile}::main
}	;# openIsis::Tspr

proc openIsis::MHL {str} {
	global env
	if [info exists env(ENCODING)] {
		set str [encoding convertfrom $env(ENCODING) $str]
	}
	# ::puts stderr "MHLing '$str'"
	regsub -all {(<[^=>]*)=[^>]+>} $str {\1>} str; # dump <a=b> substitutions
	regsub -all {><} $str {; } str; # replace >< pairs
	regsub -all {[><]} $str {} str; # nuke other ><
	regsub {^\^.} $str {} str; # kill initial subfield spec
	regsub -all {\^a} $str {; } str; # ^a -> ;
	regsub -all {\^[b-i]} $str {, } str; # ^[b-i] -> ,
	regsub -all {\^.} $str {. } str; # others -> .
	return $str
}

proc openIsis::MHLl {list} {
	set ret {}
	foreach str $list {
		lappend ret [MHL $str]
	}
	return $ret
}
