/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2001-2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */

/*
	$Id: openisistcl.c,v 1.29 2003/04/07 13:12:24 kripke Exp $
	tcl/tk binding
*/


#include "openisis.h"
#include "openisistcl.h"
#include "lcom.h"
#include "luti.h"
/* luti_getembed,
 * luti_ptrincr,
 * luti_parse_path,
 * luti_free
*/


/*
include this after the Tcl stuff for the benefit of those
who use the 150% braindead gcc 2.96
which barfs on
 declaration of `index' shadows global declaration
in generic/tclDecls.h
cause string.h declares
char *index(const char *s, int c)
*/
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
#define snprintf _snprintf
#endif

/* MMMOOO TBD
	-	encoding der db beruecks
	-	tms nach close & reopen ?
	-	r-opts & s-opts zusammenlegen
	-	wrap rqs if .data || .idx before send
*/


/*	============================ types ==================================
*/

/* name of local schema */
#define OIT_STB0  "openIsisRoot"

typedef enum {
	RC_ADD, RC_CREA, RC_DEL, RC_DO, RC_DONE, RC_FMT, RC_GET, RC_SET,
	RC_WRITE, RC_CLEAR, RC_SERI, RC_DESER
	/* MMM wrap/append Tcl_GetCommandInfo */
	/* see ROpts in OpRec */
} RCmd;

typedef enum {
	SC_ARQS, SC_ARSP, SC_CREA, SC_DEL, SC_NEW, SC_OPEN, SC_RECV, SC_RQS
	/* MMM close db ; reset ases (further rqs modifs in another sess) */
	,SC_QRY
	/* see SOpts in CmdStub */
} SCmd;

#define OIT_RS_USED      0x0100
#define OIT_RS_RQS       0x0200  /* schemas request rec */
#define OIT_RS_RSP       0x0400  /* schemas response rec */
#define OIT_RS_CFG       0x0800  /* schemas config rec */
#define OIT_RS_FRE       0x1000  /* free associated memory */
#define OIT_RS_SCM       0x0E00  /* schema mask */
#define OIT_RS_SZM       0x00FF  /* size mask */

typedef struct OITSess OITSess;

typedef struct {
	OpenIsisRec       *rec;
	OpenIsisDb        *db;   /* own db in recs, embedded db in OITCont */
	const OpenIsisFdt *fdt;  /* own fdt */
	OITSess           *ses;  /* allocator */
	const char        *cmd;  /* associated tcl ip cmd */
	int                stat;
} OITRec;

typedef struct {
	OITRec          env;
	OpenIsisStub    stb;
	OITRec        **recs;   /* embedded recs */
	int             numr;
} OITCont;

#define OIT_ST_ROOT      0x0001
#define OIT_ST_TCL       0x0002
#define OIT_ST_OINIT     0x0004

typedef struct {
	OpenIsisStub      stb;
	OITRec            cfg;  /* direct schema copy allocated by ses0 */
	OITCont          *rqs;
	OITCont          *rsp;
	OITSess          *ases; /* act session of rqs and rsp */
	OpenIsisDb       *db;   /* db last used */
	const char       *cmd;  /* associated tcl ip cmd */
	Tcl_Obj          *dfltproc;
	Tcl_Obj          *actproc;
	int               stat;
} OITStub;

struct OITSess {
	OpenIsisSession   ses;
	Tcl_Interp       *ip;
	OITStub          *astb;  /* act stub in use */
	OITRec          **recs;
	int               numr;
	int               six;
};

static OITSess *Sessions = 0;
static int      NumSessions = 0;

/*	============================ records ================================
*/

#define OIT_RECINCR   32
#define OIT_MAXRECS   65535
#define OIT_SESSINCR  1
#define OIT_MAXSESS   255

static void CtorRec (OITRec *that, OITSess *ses, int siz) {
	memset (that, 0, (unsigned)siz);
	that->stat = siz;
	that->ses  = ses;
}

static int NewRec (OITSess *ses, int siz) {
	int j;
	if (siz > OIT_RS_SZM) {
		log_msg (LOG_ERROR, "NewRec: unexpected size %d\n", siz);
		return -1;
	}
	for (j = ses->numr; 0 <= --j;  ) {
		if (! ses->recs[j]) {
			goto allcj;
		}
		if (! (OIT_RS_USED & ses->recs[j]->stat) &&
			siz == (OIT_RS_SZM & ses->recs[j]->stat)  /* may be <= */
		) {
			goto done;
		}
	}
	j = luti_ptrincr (ses->ses,
		&ses->recs, &ses->numr, OIT_RECINCR, sizeof (OITRec*), OIT_MAXRECS);
	if (0 > j) {
		return -1;
	}
allcj:
	ses->recs[j] = (OITRec*) openIsisMAlloc (ses->ses, siz);
	if (! ses->recs[j]) {
		return -1;
	}
	CtorRec (ses->recs[j], ses, siz);
done:
	ses->recs[j]->stat |= OIT_RS_USED;
	return j;
}

static int NewDbRec (OITSess *ses, OpenIsisDb *db, const OpenIsisFdt *fdt) {
	int j = NewRec (ses, sizeof (OITRec));
	if (0 <= j) {
		ses->recs[j]->db = db;
		ses->recs[j]->fdt = fdt;
	}
	return j;
}

static int NewCont (OITSess *ses, const OpenIsisFdt *fdt, int type) {
	int j = NewRec (ses, sizeof (OITCont));
	if (0 <= j) {
		ses->recs[j]->fdt = fdt;
		ses->recs[j]->stat |= type;
	}
	return j;
}

static void DtorRecs (OITSess *ses, OITRec **recs, int numr, int frmem);

static void DtorRec (OITRec *that, int frmem) {
	if (that) {
		OITCont    *con;
		OITSess    *ses;
		int         siz, type;
		ses = that->ses;
		if (that->cmd) {
			if (frmem) {
				that->stat |= OIT_RS_FRE;
			}
			Tcl_DeleteCommand (ses->ip, (char*)that->cmd);
			return;
		}
		type = OIT_RS_SCM & that->stat;
		siz = OIT_RS_SZM & that->stat;
		switch (type) {
		case 0:
			if (that->rec) {
				openIsisMFree (ses->ses, that->rec);
			}
			break;
		case OIT_RS_RQS:
			if (that->rec) {
				openIsisMFree (ses->ses, that->rec);
			}
			/* fall thru */
		case OIT_RS_RSP:
			con = (OITCont*)that;
			DtorRecs (ses, con->recs, con->numr, 0);
			break;
		default:
			log_msg (LOG_ERROR, "DtorRec: unexpected type %x\n", type);
			return;
		}
		if (frmem || (OIT_RS_FRE & that->stat)) {
			openIsisMFree (ses->ses, that);
		}
		else {
			CtorRec (that, ses, siz);
		}
	}
}

static void DtorRecs (OITSess *ses, OITRec **recs, int numr, int frmem) {
	if (recs) {
		while (0 <= --numr) {
			DtorRec (recs[numr], frmem);
		}
		openIsisMFree (ses->ses, recs);
	}
}

static int CtorSess (OpenIsisSession ses, Tcl_Interp *ip) {
	int j = luti_ptrincr (openisis_ses0,
		&Sessions, &NumSessions, OIT_SESSINCR, sizeof (OITSess), OIT_MAXSESS);
	if (0 > j) {
		return -1;
	}
	Sessions[j].ses = ses;
	Sessions[j].ip  = ip;
	Sessions[j].six = j;
	return j;
}

static void DtorSess (OITSess *that) {
	DtorRecs (that, that->recs, that->numr, !0);
}

static void ExitSess () {
	int j;
	if (NumSessions) {
		for (j = NumSessions; 0 <= --j;  ) {
			DtorSess (Sessions + j);
		}
		openIsisMFree (openisis_ses0, Sessions);
		Sessions = 0;
		NumSessions = 0;
	}
}

/*	---------------------------------------------------------------------
*/

static unsigned _RecId = 0;

static char* NewRecId (char *buf) {
	sprintf (buf, "openIsisRec%u", ++_RecId);
	return buf;
}

static void TclDelRec (ClientData cld) {
	OITRec *that = (OITRec*)cld;
	if (that->cmd) {
		openIsisMFree (that->ses->ses, (void*)that->cmd);
		that->cmd = 0;
	}
	DtorRec ((OITRec*)cld, 0);
}

static int CmdRec (
	ClientData rid, Tcl_Interp *ip, int argc, Tcl_Obj* const argv[]
);

static int CrtRecCmd (OITSess *ses, const char *name, OITRec *rec, int srst) {
	char buf[64];
	if (! name || ! *name) {
		name = NewRecId (buf);
	}
	rec->cmd = (const char*) openIsisMDup (ses->ses, name, -1);
	if (! rec->cmd) {
		Tcl_AppendResult (ses->ip, "CrtRecCmd: out of memory", 0);
		return TCL_ERROR;
	}
	Tcl_CreateObjCommand (ses->ip, (char*)name, &CmdRec, rec, &TclDelRec);
	if (srst) {
		Tcl_SetResult (ses->ip, (char*)name, TCL_VOLATILE);
	}
	return TCL_OK;
}

static char **ToArgv (
	OpenIsisSession ses, Tcl_Obj* const objv[], int objc, char* buf, int siz
) {
	char *str, *nb;
	char *res  = buf;
	int   posp = 0;
	int   posv = objc * sizeof (char*);
	int   j, len, nsz;
	for (j = 0; objc > j; ++j) {
		str = Tcl_GetStringFromObj (objv[j], &len);
		++len;
		nsz = posv + len;
		if (siz < nsz) {
			nsz *= 2;
			nb = (char*) openIsisMAlloc (ses, nsz);
			if (! nb) {
				log_msg (LOG_ERROR, "ToArgv: out of memory\n");
				return 0;
			}
			memcpy (nb, res, (unsigned) posv);
			if (res != buf) {
				openIsisMFree (ses, res);
			}
			siz = nsz;
			res = nb;
		}
		*(char**)(res + posp) = res + posv;
		strcpy (res + posv, str);
		posp += sizeof (char*);
		posv += len;
	}
	return (char**)res;
}

#define FldObj( f ) Tcl_NewStringObj( (f)->val, (f)->len )


static int BuildEmbRecs (
	OITCont *that, OpenIsisRec **recs, int numr, int frr
) {
	OITSess      *ses;
	OITRec      **oirs;
	OpenIsisDb   *db;
	OpenIsisFdt  *fdt;
	int           buf[1000];
	int          *idx = buf;
	int           j;
	ses = that->env.ses;
	oirs = (OITRec**) openIsisMAlloc (ses->ses,
		(int) (numr * sizeof (OITRec*)));
	if (! oirs) {
		if (frr) {
			luti_free (ses->ses, (void**)recs, numr);
		}
		return -1;
	}
	if (1000 < numr) {
		idx = (int*) openIsisMAlloc (ses->ses,
			(int) (numr * sizeof (int)));
		if (! idx) {
			if (frr) {
				luti_free (ses->ses, (void**)recs, numr);
			}
			openIsisMFree (ses->ses, oirs);
			return -1;
		}
	}
	db = that->env.db;
	fdt = db ? db->fdt : 0;
	for (j = numr; 0 <= --j;  ) {
		idx[j] = NewDbRec (ses, db, fdt);
		if (0 > idx[j]) {
			if (frr) {
				luti_free (ses->ses, (void**)recs, numr);
			}
			openIsisMFree (ses->ses, oirs);
			if (idx != buf) {
				openIsisMFree (ses->ses, idx);
			}
			return -1;
		}
	}
	for (j = numr; 0 <= --j;  ) {
		oirs[j] = ses->recs[idx[j]];
		oirs[j]->rec = recs[j];
	}
	that->recs = oirs;
	that->numr = numr;
	if (idx != buf) {
		openIsisMFree (ses->ses, idx);
	}
	return numr;
}

static int BuildRqsRecs (OITCont *that) {
	OpenIsisRec *recs[2] = { 0, 0 };  /* REC, IDX */
	return BuildEmbRecs (that, recs, 2, 0);
}

static int BuildRspRecs (OITCont *that) {
	OpenIsisRec **recs;
	int           numr;
	numr = openIsisNGetResult (
		that->stb, that->env.ses->ses, 0, &recs, 0);
	if (0 >= numr) {
		return numr;
	}
	return BuildEmbRecs (that, recs, numr, !0);
}

static int UsageRec (const char *arg0, Tcl_Interp *ip) {
	if (! arg0) {
		arg0 = "<openIsisRec>";
	}
	Tcl_AppendResult (ip,
		"usage: ", arg0,
		" add field value ?field value ...? |",
		" create ?options? newname ?field value ...? |",
		" del ?field ...? |",
		" do ?tagvar? valvar body |",
		" done |",
		" fmt ?options? format |",
		" get ?field ...? |",
		" set field ?value field value ...? |",
		" write |",
		" clear |",
		" .path ?option arg ...?",
		0);
	return TCL_ERROR;
}

static int OpRec (OITRec *that, Tcl_Interp *ip, const char *arg0,
	Tcl_Obj *arg1, int argc, Tcl_Obj* const argv[], int rdonly
) {
	static const/*Tcl 4 has fixed this*/ char *ROpts[] = {
		"add", "create", "delete", "do", "done", "fmt", "get", "set",
		"write", "clear", "serialize", "deserialize",
		0
	};

	char        buf[2048];
	char      **argp   =   0;
	OITSess    *ses;
	const char *opt;
	int         cmd, rt, setf = 0;

	if (! arg1) {
		return UsageRec (arg0, ip);
	}
	if (! arg0) {
		arg0 = "<openIsisRec>";
	}
	if (! that || ! (ses = that->ses)) {
		Tcl_AppendResult (ip, arg0, ": record corrupted", 0);
		return TCL_ERROR;
	}
	if (ip != ses->ip) {
		Tcl_AppendResult (ip, arg0, ": session corrupted", 0);
		return TCL_ERROR;
	}

	opt = Tcl_GetStringFromObj (arg1, 0);

	/* path to embedded rec */
	if ('.' == *opt) {
		int type = OIT_RS_SCM & that->stat;
		switch (type) {
		case OIT_RS_RQS:
			{	OITCont *con;
				OITRec  *rec;
				int      tag, occ;
				con = (OITCont*) that;
				opt = luti_parse_path (opt, openIsisComRqsPar,
					0, &tag, &occ);
				if (0 == opt ||
					0 < occ ||
					(LCOM_RQS_REC != tag && LCOM_RQS_IDX != tag)
				) {
					Tcl_AppendResult (ip, arg0,
						": no such child", 0);
					return TCL_ERROR;
				}
				if (! con->numr) {
					rt = BuildRqsRecs (con);
					if (0 > rt) {
						Tcl_AppendResult (ip, arg0,
							": child allocation failure", 0);
						return TCL_ERROR;
					}
				}
				rec = con->recs[LCOM_RQS_REC == tag ? 0 : 1];
				NewRecId (buf);
				if (*opt) {
					Tcl_Obj *obj = Tcl_NewStringObj (opt, (int) strlen (opt));
					if (obj) {
						Tcl_IncrRefCount (obj);
						rt = OpRec (rec, ip, buf, obj, argc, argv, 0);
						Tcl_DecrRefCount (obj);
						return rt;
					}
					return TCL_ERROR;
				}
				if (! argc) {
					return CrtRecCmd (ses, buf, rec, !0);
				}
				return OpRec (rec, ip, buf, argv[0], argc - 1, argv + 1, 0);
			}
		case OIT_RS_RSP:
			{	OITCont *con;
				OITRec  *rec;
				int      tag, occ;
				con = (OITCont*) that;
				opt = luti_parse_path (opt, openIsisComRspPar,
					0, &tag, &occ);
				if (0 == opt ||
					LCOM_RSP_DATA != tag
				) {
					Tcl_AppendResult (ip, arg0,
						": no such child", 0);
					return TCL_ERROR;
				}
				if (!(rt = con->numr)) {
					rt = BuildRspRecs (con);
					if (0 > rt) {
						Tcl_AppendResult (ip, arg0,
							": child allocation failure", 0);
						return TCL_ERROR;
					}
				}
				if (0 > occ) {
					occ = 0;
				}
				if (rt <= occ) {
					sprintf (buf, "%d[%d]", occ, rt);
					Tcl_AppendResult (ip, arg0,
						": no such child ", buf, 0);
					return TCL_ERROR;
				}
				rec = con->recs[occ];
				NewRecId (buf);
				if (*opt) {
					Tcl_Obj *obj = Tcl_NewStringObj (opt, (int) strlen (opt));
					if (obj) {
						Tcl_IncrRefCount (obj);
						rt = OpRec (rec, ip, buf, obj, argc, argv, 0);
						Tcl_DecrRefCount (obj);
						return rt;
					}
					return TCL_ERROR;
				}
				if (! argc) {
					return CrtRecCmd (ses, buf, rec, !0);
				}
				return OpRec (rec, ip, buf, argv[0], argc - 1, argv + 1, 0);
			}
		default:
			{	OpenIsisRec *rec;
				OITRec      *oir;
				rec = luti_getembed (
					ses->ses, that->rec, opt, that->fdt);
				if (! rec) {
					Tcl_AppendResult (ip, arg0,
						": no such child ", opt, 0);
					return TCL_ERROR;
				}
				rt = NewDbRec (ses, 0, 0);
				if (0 > rt) {
					Tcl_AppendResult (ip, "OpRec: out of memory", 0);
					return TCL_ERROR;
				}
				oir = ses->recs[rt];
				NewRecId (buf);
				if (! argc) {
					return CrtRecCmd (ses, buf, oir, !0);
				}
				return OpRec (oir, ip, buf, argv[0], argc - 1, argv + 1, 0);
			}
		}
	} /* path */

	rt = Tcl_GetIndexFromObj (ip, arg1, ROpts, "option", 0, &cmd);
	if (TCL_OK != rt) {
		return TCL_ERROR;
	}
	rt = TCL_ERROR;

	switch ((RCmd)cmd) {
	case RC_SET:
		if (0 == argc) {
			return UsageRec (arg0, ip);
		}
		if (1 == argc) {
			goto op_get;
		}
		if ( that->rec )
			setf = OPENISIS_RCHG;
		/* fall thru */
	case RC_ADD:
		if (rdonly) {
			Tcl_AppendResult (ip, arg0, ": readonly record", 0);
			return TCL_ERROR;
		}
		argp = ToArgv (ses->ses, argv, argc, buf, sizeof (buf));
		if (argp) {
			that->rec = openIsisRSet (ses->ses, that->rec,
				OPENISIS_RARGV | OPENISIS_RFDT | OPENISIS_RDIS | setf | argc,
				that->fdt, argp);
			if (that->rec) {
				rt = TCL_OK;
			}
		}
		break;
	case RC_DEL:
		if (rdonly) {
			Tcl_AppendResult (ip, arg0, ": readonly record", 0);
			return TCL_ERROR;
		}
		if (argc) {
			argp = ToArgv (ses->ses, argv, argc, buf, sizeof (buf));
			if (argp) {
				that->rec = openIsisRSet (ses->ses, that->rec,
					OPENISIS_RARGV | OPENISIS_RFDT | OPENISIS_RDEL | argc,
					that->fdt, argp);
				rt = TCL_OK;
			}
		}
		else {
			if (that->rec) {
				openIsisMFree (that->ses->ses, that->rec);
				that->rec = 0;
			}
			rt = TCL_OK;
		}
		break;
	case RC_CREA:
		break;
	case RC_DO: {
		Tcl_Obj *tagvar = 0;
		Tcl_Obj *valvar;
		Tcl_Obj *tag = 0;
		Tcl_Obj *val;
		Tcl_Obj *body;
		OpenIsisField *f;
		int i;
		if ( 2 > argc )
			return UsageRec (arg0, ip);
		body = argv[--argc];
		valvar = argv[--argc];
		if ( 1 == argc )
			tagvar = argv[0];
#ifndef DONTREUSE
		/* prepare object for var */
		if ( ! Tcl_ObjSetVar2( ip, valvar, 0,
			val = Tcl_NewObj(), TCL_LEAVE_ERR_MSG )
		)
			return TCL_ERROR;
		if ( tagvar && ! Tcl_ObjSetVar2( ip, tagvar, 0,
			tag = Tcl_NewIntObj(0), TCL_LEAVE_ERR_MSG )
		)
			return TCL_ERROR;
#endif

		/* go loop */
		for ( i = that->rec->len, f = that->rec->field; i--; f++ ) {
#ifndef DONTREUSE
			/*
				if somebody shares the object, we have to set a new one
				(or Tcl_SetStringObj will panic).
				if it's not owned by the var, it's either owned by someone else
				or nobody, i.e. deleted
			*/
			if ( val == Tcl_ObjGetVar2( ip, valvar, 0, TCL_LEAVE_ERR_MSG )
				&& !Tcl_IsShared( val )
			)
				Tcl_SetStringObj( val, (char*)f->val, f->len );
			else
#endif
				if ( !Tcl_ObjSetVar2( ip, valvar, 0,
					val = FldObj( f ), TCL_LEAVE_ERR_MSG )
				)
					return TCL_ERROR;
			if ( tagvar ) {
#ifndef DONTREUSE
				if ( tag == Tcl_ObjGetVar2( ip, tagvar, 0, TCL_LEAVE_ERR_MSG )
					&& !Tcl_IsShared( tag )
				)
					Tcl_SetIntObj( tag, f->tag );
				else
#endif
					if ( !Tcl_ObjSetVar2( ip, tagvar, 0,
						tag = Tcl_NewIntObj( f->tag ), TCL_LEAVE_ERR_MSG )
					)
						return TCL_ERROR;
			}	/* tagvar */
			switch (rt = Tcl_EvalObjEx( ip, body, 0 )) {
			case TCL_CONTINUE:
				rt = TCL_OK;
			case TCL_OK:
				continue;
			case TCL_BREAK:
				rt = TCL_OK;
			case TCL_RETURN:
			case TCL_ERROR:
			default:
				goto done;
			}
		}
		break;
	} /* case RC_DO */
	case RC_DONE:
		if (! that->cmd) {
			Tcl_AppendResult (ip, arg0, ": no command bound to rec", 0);
			return TCL_ERROR;
		}
		if (strcmp (that->cmd, arg0)) {
			Tcl_AppendResult (ip, arg0, ": command mismatch: ", that->cmd, 0);
			return TCL_ERROR;
		}
		if (rdonly) {
			Tcl_AppendResult (ip, arg0, ": readonly record", 0);
			return TCL_ERROR;
		}
		Tcl_DeleteCommand (ip, (char*)arg0);
		return TCL_OK;
	case RC_FMT:
		break;
	case RC_GET:
	op_get:
		{
			OpenIsisField *fld;
			const char *path, *rem;
			int tag, occ, i, j;
			Tcl_Obj *list;

			Tcl_ResetResult (ip);
			if (! that->rec) {
				return TCL_OK;
			}
			list = Tcl_NewListObj( 0, 0 );
			if (! argc) {
				for ( i = that->rec->len, fld = that->rec->field; i--; fld++ )
					Tcl_ListObjAppendElement( ip, list, FldObj(fld) );
			}
			else {
				for (j = 0; argc > j; ++j) {
					path = Tcl_GetStringFromObj (argv[j], 0);
					rem = luti_parse_path (path, that->fdt, 0, &tag, &occ);
					if (! rem || *rem) {
						Tcl_ResetResult (ip);
						Tcl_AppendResult (ip, arg0,
							": no such path: ", path, 0);
						goto geterr;
					}
					if ( 0 > occ ) {
						for ( i = that->rec->len, fld = that->rec->field; i--; fld++ )
							if ( tag == fld->tag )
								Tcl_ListObjAppendElement( ip, list, FldObj(fld) );
					} else { /* specific occ wanted */
						fld = openIsisROccurence (that->rec, tag, occ);
						if (! fld) {
							Tcl_ResetResult (ip);
							Tcl_AppendResult (ip, arg0,
								": no such field: ", path, 0);
							goto geterr;
						}
						if ( TCL_OK != Tcl_ListObjAppendElement( ip, list, FldObj(fld) ) ) {
							goto geterr;
						}
					}
				}
			}
			Tcl_SetObjResult( ip, list );
			return TCL_OK;
		geterr:
			Tcl_DecrRefCount( list ); /* free */
			return TCL_ERROR;
		} /* RC_GET */
	case RC_WRITE: {
			if ( that->rec) {
				int ok = openIsisDWrite( that->ses->ses, that->rec->dbid, that->rec );
				openIsisSMsg( that->ses->ses, LOG_ERROR,
					"write db %d = %d mfn %d", that->rec->dbid, ok, that->rec->rowid );
			}
			return TCL_OK;
		} /* RC_WRITE */
	case RC_CLEAR: {
			if ( that->rec) {
				OPENISIS_CLRREC( that->rec );
			}
			return TCL_OK;
		} /* RC_CLEAR */
	case RC_SERI: {
		char *b;
		int len;

		if (! that->rec) {
			Tcl_ResetResult (ip);
			return TCL_OK;
		}
		b = that->rec->used <= (int)sizeof(buf)
			? buf : Tcl_Alloc( that->rec->used+1 );
		len = openIsisRSerialize( b, that->rec );
		/*
		faster, but not allowed :(
		Tcl_Obj *ret = Tcl_NewStringObj(0,0);
		Tcl_SetObjLength( ret, that->rec->used );
		Tcl_SetObjLength( ret, openIsisRSerialize( ret->bytes, that->rec ) );
		*/
		/* do NOT include the final blankline */
		Tcl_SetObjResult( ip, Tcl_NewStringObj( b, len-1 ) );
		if ( buf != b )
			Tcl_Free( b );
		return TCL_OK;
	} /* RC_SERI */
	case RC_DESER: {
		char *b;
		int len;

		if ( 1 != argc )
			return UsageRec (arg0, ip);
		b = Tcl_GetStringFromObj( argv[0], &len );
		if ( ! b )
			Tcl_ResetResult (ip);
		else {
			int ret = openIsisRDeserialize( ses->ses, &that->rec,
				b, len, OPENISIS_RDIS|OPENISIS_STOPONEMPTY );
			Tcl_SetObjResult( ip, Tcl_NewIntObj( ret ) );
		}

		return TCL_OK;
	} /* RC_DESER */
	default:
		Tcl_AppendResult (ip, arg0, ": unrecognized command ", opt, 0);
	}	/* switch */

done:
	if (argp && (char**)buf != argp) {
		openIsisMFree (ses->ses, argp);
	}
	return rt;
}

static int CmdRec (
	ClientData cld, Tcl_Interp *ip, int argc, Tcl_Obj* const argv[]
) {
	const char *arg0 = 0 >= argc ? 0 : Tcl_GetStringFromObj (argv[0], 0);
	if (1 >= argc) {
		return UsageRec (arg0, ip);
	}
	return OpRec ((OITRec*)cld, ip, arg0, argv[1], argc - 2, argv + 2, 0);
}

/*	---------------------------------------------------------------------
*/

static int RspCb (
	void *cld, OpenIsisStub stb, OpenIsisRec *rsp, OpenIsisDb *db
) {
	OITStub  *that = (OITStub*) cld;
	int       rt;

	if (! that) {
		log_msg (LOG_ERROR, "RspCb: rsp w/o stb\n");
		return -1;
	}
	if (stb != that->stb) {
		log_msg (LOG_ERROR, "RspCb: stb changed\n");
		return -1;
	}
	if (! that->ases) {
		log_msg (LOG_ERROR, "RspCb: rsp w/o ses\n");
		return -1;
	}
	if (! that->rqs) {
		log_msg (LOG_ERROR, "RspCb: rsp w/o rqs\n");
		return -1;
	}
	if (that->rsp) {
		log_msg (LOG_WARN, "RspCb: multiple rsp\n");
		DtorRec ((OITRec*)that->rsp, 0);
		that->rsp = 0;
	}

	rt = NewCont (that->ases, openIsisComRspPar, OIT_RS_RSP);
	if (0 > rt) {
		log_msg (LOG_WARN, "RspCb: out of memory\n");
		return -1;
	}
	that->rsp = (OITCont*) that->ases->recs[rt];
	that->rsp->env.rec = rsp;
	that->rsp->stb = that->stb;
	that->db = db;

	rt = 0;
	if (that->actproc) {
		rt = Tcl_EvalObj (that->ases->ip, that->actproc);
	}
	else if (that->dfltproc) {
		rt = Tcl_EvalObj (that->ases->ip, that->dfltproc);
	}
	return rt;
}

static void StbDelCb (void *cld, OpenIsisStub stb, void *cbd);
static void TclDelStb (ClientData cld);
static int CmdStub (
	ClientData cld, Tcl_Interp *ip, int argc, Tcl_Obj* const argv[]
);

static OITStub* CtorStub (Tcl_Interp *ip,
	const char *name, int argc, const char **argv, const char *proc
) {
	OpenIsisStubCbData  scd;
	OpenIsisSchema     *sch;
	OITStub            *that;
	int                 stat  =  0;

	if (! name) {
		name = OIT_STB0;
		stat = OIT_ST_ROOT;
	}
	if (! NumSessions) {
		log_msg (LOG_ERROR,
			"CtorStub(%s): no root session\n", name);
		return 0;
	}
	if (ip != Sessions->ip) {
		log_msg (LOG_ERROR,
			"CtorStub(%s): must not create stub in derived session\n", name);
		return 0;
	}
	that = (OITStub*) openIsisMAlloc (openisis_ses0, sizeof (OITStub));
	if (! that) {
		return 0;
	}
	that->cmd = openIsisMDup (openisis_ses0, name, -1);
	if (! that->cmd) {
		openIsisMFree (openisis_ses0, that);
		return 0;
	}

	memset (&scd, 0, sizeof (OpenIsisStubCbData));
	scd.dfltcb = &RspCb;
	scd.delcb  = &StbDelCb;
	scd.dfltcld = scd.delcld = that;

	if (OIT_ST_ROOT & stat) {
		that->stb = openIsisNInit (argc, argv, &scd);
	}
	else {
		that->stb = openIsisNOpen (name, argc, argv, &scd);
	}
	if (! that->stb) {
		openIsisMFree (openisis_ses0, (void*)that->cmd);
		openIsisMFree (openisis_ses0, that);
		return 0;
	}

	sch = openIsisNSchema (that->stb);
	that->cfg.rec = sch->cfg;
	that->cfg.fdt = openIsisSyspar;
	that->cfg.ses = Sessions;
	that->cfg.stat = OIT_RS_USED | OIT_RS_CFG;

	if (proc) {
		that->dfltproc = Tcl_NewStringObj (proc, (int) strlen (proc));
		if (that->dfltproc) {
			Tcl_IncrRefCount (that->dfltproc);
		}
	}

	that->stat = stat | OIT_ST_TCL | OIT_ST_OINIT;

	Tcl_CreateObjCommand (ip, (char*)that->cmd, &CmdStub, that, &TclDelStb);
	Tcl_SetResult (ip, (char*)that->cmd, TCL_VOLATILE);
	return that;
}

static void DtorStub (OITStub *that, int where) {
	const char *cmd = that->cmd;
	if (! cmd) {
		cmd = "<<NULL>>";
	}
	if (! NumSessions) {
		log_msg (LOG_ERROR, "DtorStub(%s): no root session\n", cmd);
		return;
	}
	that->stat &= ~where;
	if (OIT_ST_TCL & that->stat) {
		if (! that->cmd) {
			log_msg (LOG_ERROR, "DtorStub(%s): no command\n", cmd);
			return;
		}
		Tcl_DeleteCommand (Sessions->ip, (char*)that->cmd);
		return;
	}
	if (OIT_ST_OINIT & that->stat) {
		if (OIT_ST_ROOT & that->stat) {
			openIsisNDeinit ();
		}
		else {
			openIsisNClose (that->stb);
		}
		return;
	}
	if (that->dfltproc) {
		Tcl_DecrRefCount (that->dfltproc);
	}
	if (that->actproc) {
		Tcl_DecrRefCount (that->actproc);
	}
	if (that->rqs) {
		DtorRec ((OITRec*)that->rqs, 0);
	}
	if (that->rsp) {
		DtorRec ((OITRec*)that->rsp, 0);
	}
	if (that->cmd) {
		openIsisMFree (openisis_ses0, (void*)that->cmd);
	}
	openIsisMFree (openisis_ses0, that);
}

static void TclDelStb (ClientData cld) {
	DtorStub ((OITStub*)cld, OIT_ST_TCL);
}

static void StbDelCb (void *cld, OpenIsisStub stb, void *cbd) {
	(void) stb;
	if (cbd) {
		if (((OITStub*)cbd)->actproc) {
			Tcl_DecrRefCount (((OITStub*)cbd)->actproc);
			((OITStub*)cbd)->actproc = 0;
		}
		return;
	}
	DtorStub ((OITStub*)cld, OIT_ST_OINIT);
}

static int UsageStub (const char *argv0, Tcl_Interp *ip) {
	Tcl_AppendResult (ip,
		"usage: ",
		(argv0 ? argv0 : "<openIsisStub>"),
		" .req |",
		" .res ?option ...? |",
		" create newname ?-cfg val? |",
		" delete |",
		" new ?-db db? ?name? |",
		" open ?-cfg val ...? |",
		" recv |",
		" req ?-param val?",
		0);
	return TCL_ERROR;
}

static int CmdStub (
	ClientData cld, Tcl_Interp *ip, int argc, Tcl_Obj* const argv[]
) {
	static const/*Tcl 4 has fixed this*/ char *SOpts[] = {
		".req", ".res", "create", "delete", "new", "open", "recv", "req",
		"qry",/*KKK should be a req, but couldn't figure out */ 0
		/* MMM ROpts und SOpts disjunkt */
	};

	char         buf[2048];
	OITStub     *that   =   (OITStub*) cld;
	OITSess     *ses;
	char       **argp   =   0;
	const char  *arg0;
	int          cmd, rt;

	if (1 > argc) {
		return TCL_ERROR;
	}
	arg0 = Tcl_GetStringFromObj (argv[0], 0);
	if (2 > argc) {
		return UsageStub (arg0, ip);
	}
	/* MMM ip != ases/Session0->ip je nach cmd */
	/* MMM cmd release session ? auto on .req delete all ? auto on ip chg? */
	if (!(ses = that->ases)) {
		if (! NumSessions) { 
			Tcl_AppendResult (ip, arg0, ": no session", 0);
			return TCL_ERROR;
		}
		ses = that->ases = Sessions;
	}
	if (ip != ses->ip) {
		Tcl_AppendResult (ip, arg0, ": session changed", 0);
		return TCL_ERROR;
	}

	rt = Tcl_GetIndexFromObj (ip, argv[1], SOpts, "option", 0, &cmd);
	if (TCL_OK != rt) {
		OpenIsisSchema *sch;
		rt = CmdRec (&that->cfg, ip, argc, argv);
		sch = openIsisNSchema (that->stb);
		sch->cfg = that->cfg.rec;
		return rt;
	}
	rt = TCL_ERROR;

	switch ((SCmd)cmd) {
	case SC_ARQS:
		{
			/* MMM crt rec cmd for rqs ? */
			if (3 > argc) {
				return UsageRec (0, ip);
			}
			if (! that->rqs) {
				rt = NewCont (ses, openIsisComRqsPar, OIT_RS_RQS);
				if (0 > rt) {
					Tcl_AppendResult (ip, arg0, ": out of memory", 0);
					return TCL_ERROR;
				}
				that->rqs = (OITCont*) ses->recs[rt];
			}
			rt = OpRec ((OITRec*)that->rqs, ip,
				0, argv[2], argc - 3, argv + 3, 0);
			return rt;
		}

	case SC_ARSP:
		{
			if (3 > argc) {
				Tcl_SetResult (ip,
					that->rsp && that->rsp->env.rec ? "1" : "0",
					TCL_STATIC);
				return TCL_OK;
			}
			if (! that->rsp || ! that->rsp->env.rec) {
				Tcl_AppendResult (ip, arg0, ": no response available", 0);
				return TCL_ERROR;
			}
			rt = OpRec ((OITRec*)that->rsp, ip,
				0, argv[2], argc - 3, argv + 3, !0);
			return rt;
		}

	case SC_CREA:
		{
			OITStub *news;
			if (3 > argc) {
				return UsageStub (arg0, ip);
			}
			argp = ToArgv (openisis_ses0,
				argv + 2, argc - 2, buf, sizeof (buf));
			if (argp) {
				if (! *argp[0]) {
					if (argp != (char**)buf) {
						openIsisMFree (openisis_ses0, argp);
					}
					return UsageStub (arg0, ip);
				}
				news = CtorStub (ip, argp[0], argc - 3,
					(const char**) argp + 1, 0);
				if (argp != (char**)buf) {
					openIsisMFree (openisis_ses0, argp);
				}
				if (! news) {
					Tcl_AppendResult (ip, arg0, ": ", argp[0],
						": illegal schema or out of memory", 0);
					return TCL_ERROR;
				}
				return TCL_OK;
			}
			return TCL_ERROR;
		}

	case SC_DEL:
		Tcl_DeleteCommand (ip, (char*)arg0);
		return TCL_OK;

	case SC_NEW:
		{
			OpenIsisDb *db   = that->db;
			char       *name = 0;
			if (2 < argc) {
				if (5 < argc) {
					return UsageStub (arg0, ip);
				}
				argp = ToArgv (ses->ses,
					argv + 2, argc - 2, buf, sizeof (buf));
				if (3 == argc) {
					name = argp[0];
				}
				else {
					if (strcmp (argp[0], "-db")) {
						if (argp != (char**)buf) {
							openIsisMFree (ses->ses, argp);
						}
						return UsageStub (arg0, ip);
					}
					db = openIsisNDbByName (that->stb, argp[1]);
					if (! db) {
						Tcl_AppendResult (ip, arg0, ": no such db: ",
							argp[1], 0);
						if (argp != (char**)buf) {
							openIsisMFree (ses->ses, argp);
						}
						return TCL_ERROR;
					}
					if (5 == argc) {
						name = argp[2];
					}
				}
			}
			if (! db) {
				Tcl_AppendResult (ip, arg0, ": no db specified", 0);
				return TCL_ERROR;
			}
			rt = NewDbRec (ses, db, db->fdt);
			if (0 > rt) {
				if (argp && argp != (char**)buf) {
					openIsisMFree (ses->ses, argp);
				}
				Tcl_AppendResult (ip, arg0, ": out of memory", 0);
				return TCL_ERROR;
			}
			rt = CrtRecCmd (ses, name, ses->recs[rt], !0);
			if (argp && argp != (char**)buf) {
				openIsisMFree (ses->ses, argp);
			}
			return rt;
		}

	case SC_OPEN:
		break;

	case SC_RECV:
		{
			if (that->rsp && that->rsp->env.rec) {
				return TCL_OK;
			}
			/* MMM TBD async mode on remote invocation */
			return TCL_OK;
		}

	case SC_RQS:
		{
			if (that->rsp) {
				DtorRec ((OITRec*) that->rsp, 0);
			}
			if (! that->rqs) {
				rt = NewCont (ses, openIsisComRqsPar, OIT_RS_RQS);
				if (0 > rt) {
					Tcl_AppendResult (ip, arg0, ": out of memory", 0);
					return TCL_ERROR;
				}
				that->rqs = (OITCont*) ses->recs[rt];
			}
			if (2 < argc) {
				Tcl_Obj *obj = Tcl_NewStringObj ("set", 3);
				if (! obj) {
					Tcl_AppendResult (ip, arg0, ": out of memory", 0);
					return TCL_ERROR;
				}
				Tcl_IncrRefCount (obj);
				rt = OpRec ((OITRec*)that->rqs, ip,
					0, obj, argc - 2, argv + 2, 0);
				Tcl_DecrRefCount (obj);
				if (TCL_OK != rt) {
					return rt;
				}
			}
			rt = openIsisNSend (that->stb, that->ases->ses,
				that->rqs->env.rec, 0, 0, !0);
			if (0 != rt) {
				openIsisI2a (buf, rt);
				if (TCL_ERROR != rt) {
					rt = TCL_ERROR;
				}
				Tcl_AppendResult (ip, arg0, ": error ", buf,
					" sending request", 0);
			}
			return rt;
		} /* case SC_RQS: */
	case SC_QRY: {
		OpenIsisSet set;
		const char *key;
		Tcl_Obj *list;
		int i;

		if (3 > argc) {
			return UsageStub (arg0, ip);
		}
		set.len = 0;
		key = Tcl_GetStringFromObj (argv[2], 0);
		i = openIsisDQuery( that->ases->ses, &set, that->db->dbid, key,
			OPENISIS_QRY_SIMPLE, 0 );
		if ( 0 > i ) {
			Tcl_AppendResult (ip, arg0, ": error on qry ", key, 0 );
			return TCL_ERROR;
		}
		list = Tcl_NewListObj( 0, 0 );
		for ( i=0; i<set.len; i++ )
			Tcl_ListObjAppendElement( ip, list, Tcl_NewIntObj(set.id[i]) );
		Tcl_SetObjResult( ip, list );
		return TCL_OK;
	} /* case SC_QRY: */
	}

	if (argp && argp != (char**)buf) {
		openIsisMFree (openisis_ses0, argp);
	}
	Tcl_AppendResult (ip, arg0, ": sorry, option not implemented yet");
	return TCL_ERROR;
}

static void AtExit (ClientData cld) {
	(void)cld;
	openIsisNDeinit ();
	ExitSess ();
}

static int CmdInit (
	ClientData cld, Tcl_Interp *ip, int argc, const char *argv[]
) {
	OITStub     *news;
	const char  *proc  =  0;
	int          j, len;

	(void)cld;
	if (! NumSessions) {
		CtorSess (OPENISIS_SES0 (), ip);
	}
	if (openisis_stub0) {
		Tcl_CmdInfo info;
		if (Tcl_GetCommandInfo (ip, OIT_STB0, &info)) {
			Tcl_SetResult (ip, OIT_STB0, TCL_STATIC);
			return TCL_OK;
		}
	}

	for (j = 1; argc > j; ++j) {
		if (! argv[j] || ! (len = strlen (argv[j]))) {
			goto usage;
		}
		if ('-' == *argv[j]) {
			if (! strncmp ("-async", argv[j], (unsigned) len)) {
				if (argc <= ++j) {
					goto usage;
				}
				proc = argv[j];
				continue;
			}
		}
		break;
	}

	/* openIsisNInit can be called multiple times */
	news = CtorStub (ip, 0, argc - j, argv + j, proc);
	if (! news) {
		Tcl_AppendResult (ip, OIT_STB0, ": out of memory", 0);
		return TCL_ERROR;
	}
	return TCL_OK;

usage:
	Tcl_AppendResult (ip,
		"usage: ", argv[0], " ?-async <cb>? ?options?", 0);
	return TCL_ERROR;
}

static int CmdOIR (
	ClientData cld, Tcl_Interp *ip, int argc, Tcl_Obj* const argv[]
) {
	OITSess      *ses;
	OITRec       *rec;
	const char   *name;
	char          buf[64];
	int           rt;

	(void)cld;
	if (! NumSessions) {
		CtorSess (OPENISIS_SES0 (), ip);
	}
	ses = Sessions;

	if (1 < argc) {
		name = Tcl_GetStringFromObj (argv[1], 0);
	}
	else {
		name = NewRecId (buf);
	}

	rt = NewDbRec (ses, 0, 0);
	if (0 > rt) {
		Tcl_AppendResult (ip, Tcl_GetStringFromObj (argv[0], 0),
			": out of memory", 0);
		return TCL_ERROR;
	}
	rec = ses->recs[rt];

	if (2 < argc) {
		rt = OpRec (rec, ip, name, argv[2], argc - 3, argv + 3, 0);
		if (TCL_OK != rt) {
			return rt;
		}
	}

	rt = CrtRecCmd (ses, name, rec, !0);
	return rt;
}

static void AddCmds (Tcl_Interp *ip, int root) {
	Tcl_CreateCommand (ip, "openIsis",
		(Tcl_CmdProc*)CmdInit, 0, root ? &AtExit : 0);
	Tcl_CreateObjCommand (ip, "openIsisRec", &CmdOIR, 0, 0);
}

/* ************************************************************
	public functions
*/

int openIsisTclNewSession (OpenIsisSession ses, Tcl_Interp *ip) {
	int id = CtorSess (ses, ip);
	if (0 <= id) {
		AddCmds (ip, 0);
	}
	return id;
}

int openIsisTclGetSession (int sid, OpenIsisSession *ses, Tcl_Interp **ip) {
	if (0 > sid || NumSessions <= sid) {
		return 0;
	}
	if (ses) {
		*ses = Sessions[sid].ses;
	}
	if (ip) {
		*ip = Sessions[sid].ip;
	}
	return !0;
}

/*
void openIsisTclDelSession (int sid) {
	if (! sid) {
		ExitSess ();
		return;
	}
	if (0 < sid && NumSessions > sid) {
		DtorSess (Sessions + sid);
	}
}
*/


int openIsisTclCreateRecCmd ( int sid, char *nam, int flg )
{
	OITSess *ses = Sessions+sid;
	int rid;
	(void)flg; /* TODO: set readonly */
	rid = NewDbRec( ses, 0, 0 );
	if ( 0 <= rid )
		CrtRecCmd( ses, nam, ses->recs[rid], 0 );
	return rid;
}	/* openIsisTclCreateRecCmd */


int openIsisTclEval ( int sid, OpenIsisSession oses,
	int cnt, const int *ids, OpenIsisRec **recs, char *script )
{
	OITSess *ses = Sessions+sid;
	int ret, i;

	ses->ses = oses;
	if ( recs )
		for ( i=cnt; i--; )
			ses->recs[ ids[i] ]->rec = recs[i];
	ret = Tcl_Eval( ses->ip, script );
	if ( recs )
		for ( i=cnt; i--; )
			recs[i] = ses->recs[ ids[i] ]->rec;
	return ret;
}	/* openIsisTclEval */


int
OpenisisTcl_Init ( Tcl_Interp *ip )
{
	AddCmds (ip, !0);
	Tcl_CreateExitHandler (&AtExit, 0);
	return TCL_OK;
}
