#!/opt/openisis/tcl/tclbuild/tclsh
#
#	openisis - an open implementation of the ISIS database
#	Version 0.8.x (microversion see file Version)
#	Copyright (C) 2003 by Erik Grziwotz, erik@openisis.org
#
#	This library is free software; you can redistribute it and/or
#	modify it under the terms of the GNU Lesser General Public
#	License as published by the Free Software Foundation; either
#	version 2.1 of the License, or (at your option) any later version.
#
#	This library is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#	Lesser General Public License for more details.
#
#	You should have received a copy of the GNU Lesser General Public
#	License along with this library; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	This software is dedicated to the memory of Eckart Dietrich.
#
#	This software is inspired by (but contains no code of) the iAPI
#	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
#	See iAPI.txt for what it contains.
#
#	$Id: test1.tcl,v 1.9 2003/03/16 19:02:51 mawag Exp $
#	test script for tcl-binding of openisis

proc putt msg {
	global prnout
	if $prnout {
		puts $msg
	}
}

proc prnres {msg stb} {
	if [$stb .res] {
		set res [$stb .res get sid ser error error2]
		set tms ""
		set dta ""
		catch { set tms [$stb .res get tms] }
		catch { set dta [$stb .res get data] }
		putt "\[$stb/[lindex $res 0]] $msg:\
			ser [lindex $res 1],\
			err [lindex $res 2]/[lindex $res 3],\
			tms $tms,\
			dta $dta"
	} else {
		putt "\[$stb] $msg: no response available"
	}
}

proc maxrow {stb db wrn} {
	$stb .req delete
	$stb req type maxrow db $db
	prnres "maxrow $db" $stb
	set row 0
	if [catch {
		set row [$stb .res get rowid]
	}] {
		if $wrn {
			global errorInfo
			putt $errorInfo
		}
	} {
		putt "maxrow = $row"
	}
	return $row
}

switch $argc {
0 { set n 1; set prnout 1 }
1 { set n [lindex $argv 0]; set prnout 0 }
default { set n [lindex $argv 0]; set prnout [lindex $argv 1] }
}

for {} { 0 != $n } { incr n -1 } {
	set stb0 [openIsis]
	putt "$stb0 created."

	$stb0 add \
		syspath /opt/openisis/db \
		db cds \
		dbpath cds \
		schema ^snonno^htonno^p4343

	set flds [ $stb0 get ]
	putt $flds

	if [catch {
		set stb1 [$stb0 create nonno]
		putt $stb1
		$stb1 dele
	}] {
		global errorInfo
		putt $errorInfo
		set stb1 ""
	}

	$stb0 .req set type open
	$stb0 req db cds
	prnres "open cds" $stb0

	maxrow $stb0 cds 1

	$stb0 .req delete
	$stb0 req type close db cds
	prnres "close cds" $stb0

	maxrow $stb0 cds 0

	$stb0 .req delete
	$stb0 req type open db cds
	prnres "open cds" $stb0

	maxrow $stb0 cds 1

	$stb0 .req delete
	$stb0 req type close db cds
	prnres "close cds" $stb0

	$stb0 .req delete
	$stb0 req type open db cds
	prnres "open cds" $stb0

	set row [maxrow $stb0 cds 1]

	for {set j 1} {$j <= $row} {incr j} {
		$stb0 .req delete
		$stb0 req type read db cds rowid $j
		set ttl ""
		catch { set ttl [$stb0 .res .data get 24] }
		if {60 < [string length $ttl]} {
			set ttl "[string range $ttl 0 57] ..."
		}
		putt "row [$stb0 .res get rowid] $ttl"
	}

	# set flds [ $stb0 .res get ]
	# puts $flds

	$stb0 dele
	putt "$stb0 deleted."

	if {! $prnout && 0 == [expr $n % 50]} {
		puts "$n ..."
	}
}

puts "ok."

